# --------------------------------------------------------------------------
# ⚠️ WARNING - AUTO-GENERATED CODE - DO NOT EDIT ⚠️
# ⚙️ Generated by 'python -m opgen'
# --------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation.
# Licensed under the MIT License.
# --------------------------------------------------------------------------
# pylint: disable=W0221,W0222,R0901,W0237
# mypy: disable-error-code=override
# ruff: noqa: E741, D402
# --------------------------------------------------------------------------

from __future__ import annotations

from typing import Optional, Sequence, Tuple, TypeVar, Union

from onnx import GraphProto, TensorProto
from onnx.defs import get_schema
from typing_extensions import TypeAlias

from onnxscript.onnx_opset._impl.opset8 import Opset8
from onnxscript.onnx_types import (
    BOOL,
    COMPLEX64,
    COMPLEX128,
    DOUBLE,
    FLOAT,
    FLOAT16,
    INT8,
    INT16,
    INT32,
    INT64,
    STRING,
    UINT8,
    UINT16,
    UINT32,
    UINT64,
)
from onnxscript.values import Op, Opset


class Opset9(Opset8):
    def __new__(cls):
        return Opset.__new__(cls, "", 9)

    T_Acosh = TypeVar("T_Acosh", DOUBLE, FLOAT, FLOAT16)

    def Acosh(self, input: T_Acosh) -> T_Acosh:
        r"""[🌐 Acosh(9)](https://onnx.ai/onnx/operators/onnx__Acosh.html#acosh-9 "Online Documentation")


        Calculates the hyperbolic arccosine of the given input tensor element-wise.


        Args:
            input: (differentiable) Input tensor
        """

        schema = get_schema("Acosh", 9, "")
        op = Op(self, "Acosh", schema)
        return op(*self._prepare_inputs(schema, input))

    T_Asinh = TypeVar("T_Asinh", DOUBLE, FLOAT, FLOAT16)

    def Asinh(self, input: T_Asinh) -> T_Asinh:
        r"""[🌐 Asinh(9)](https://onnx.ai/onnx/operators/onnx__Asinh.html#asinh-9 "Online Documentation")


        Calculates the hyperbolic arcsine of the given input tensor element-wise.


        Args:
            input: (differentiable) Input tensor
        """

        schema = get_schema("Asinh", 9, "")
        op = Op(self, "Asinh", schema)
        return op(*self._prepare_inputs(schema, input))

    T_Atanh = TypeVar("T_Atanh", DOUBLE, FLOAT, FLOAT16)

    def Atanh(self, input: T_Atanh) -> T_Atanh:
        r"""[🌐 Atanh(9)](https://onnx.ai/onnx/operators/onnx__Atanh.html#atanh-9 "Online Documentation")


        Calculates the hyperbolic arctangent of the given input tensor element-wise.


        Args:
            input: (differentiable) Input tensor
        """

        schema = get_schema("Atanh", 9, "")
        op = Op(self, "Atanh", schema)
        return op(*self._prepare_inputs(schema, input))

    T_BatchNormalization = TypeVar("T_BatchNormalization", DOUBLE, FLOAT, FLOAT16)

    def BatchNormalization(
        self,
        X: T_BatchNormalization,
        scale: T_BatchNormalization,
        B: T_BatchNormalization,
        mean: T_BatchNormalization,
        var: T_BatchNormalization,
        *,
        epsilon: float = 9.999999747378752e-06,
        momentum: float = 0.8999999761581421,
    ) -> Tuple[
        T_BatchNormalization,
        T_BatchNormalization,
        T_BatchNormalization,
        T_BatchNormalization,
        T_BatchNormalization,
    ]:
        r"""[🌐 BatchNormalization(9)](https://onnx.ai/onnx/operators/onnx__BatchNormalization.html#batchnormalization-9 "Online Documentation")


        Carries out batch normalization as described in the paper
        https://arxiv.org/abs/1502.03167. Depending on the mode it is being run,
        there are multiple cases for the number of outputs, which we list below:

        Output case #1: Y, mean, var, saved_mean, saved_var (training mode)
        Output case #2: Y (test mode)

        For previous (depreciated) non-spatial cases, implementors are suggested
        to flatten the input shape to (N x C*D1*D2 ..*Dn) before a BatchNormalization Op.
        This operator has **optional** inputs/outputs. See `ONNX <https://github.com/onnx/onnx/blob/master/docs/IR.md>`_ for more details about the representation of optional arguments. An empty string may be used in the place of an actual argument's name to indicate a missing argument. Trailing optional arguments (those not followed by an argument that is present) may also be simply omitted.


        Args:
            X: (differentiable) Input data tensor from the previous operator; dimensions
                are in the form of (N x C x D1 x D2 ... Dn), where N is the batch size,
                C is the number of channels. Statistics are computed for every channel
                of C over N and D1 to Dn dimensions. For image data, input dimensions
                become (N x C x H x W). The op also accepts single dimension input of
                size N in which case C is assumed to be 1

            scale: (differentiable) Scale tensor of shape (C).

            B: (differentiable) Bias tensor of shape (C).

            mean: (differentiable) running (training) or estimated (testing) mean tensor
                of shape (C).

            var: (differentiable) running (training) or estimated (testing) variance
                tensor of shape (C).

            epsilon: The epsilon value to use to avoid division by zero.

            momentum: Factor used in computing the running mean and variance.e.g.,
                running_mean = running_mean * momentum + mean * (1 - momentum).
        """

        schema = get_schema("BatchNormalization", 9, "")
        op = Op(self, "BatchNormalization", schema)
        return op(
            *self._prepare_inputs(schema, X, scale, B, mean, var),
            epsilon=epsilon,
            momentum=momentum,
        )

    T1_Cast = TypeVar(
        "T1_Cast",
        BOOL,
        DOUBLE,
        FLOAT,
        FLOAT16,
        INT16,
        INT32,
        INT64,
        INT8,
        STRING,
        UINT16,
        UINT32,
        UINT64,
        UINT8,
    )

    T2_Cast: TypeAlias = Union[
        BOOL,
        DOUBLE,
        FLOAT,
        FLOAT16,
        INT16,
        INT32,
        INT64,
        INT8,
        STRING,
        UINT16,
        UINT32,
        UINT64,
        UINT8,
    ]

    def Cast(self, input: T1_Cast, *, to: int) -> T2_Cast:
        r"""[🌐 Cast(9)](https://onnx.ai/onnx/operators/onnx__Cast.html#cast-9 "Online Documentation")


        The operator casts the elements of a given input tensor to a data type
        specified by the 'to' argument and returns an output tensor of the same size in
        the converted type. The 'to' argument must be one of the data types specified
        in the 'DataType' enum field in the TensorProto message.

        Casting from string tensor in plain (e.g., "3.14" and "1000") and scientific numeric representations
        (e.g., "1e-5" and "1E8") to float types is supported. For example, converting string "100.5" to an integer may
        yield result 100. There are some string literals reserved for special floating-point values;
        "+INF" (and "INF"), "-INF", and "NaN" are positive infinity, negative infinity, and not-a-number, respectively.
        Any string which can exactly match "+INF" in a case-insensitive way would be mapped to positive infinite. Similarly,
        this case-insensitive rule is applied to "INF" and "NaN". When casting from numeric tensors
        to string tensors, plain floating-point representation (such as "314.15926") would be used.
        Converting non-numerical-literal string such as "Hello World!" is an undefined behavior. Cases
        of converting string representing floating-point arithmetic value, such as "2.718", to INT is an undefined behavior.

        Conversion from a numerical type to any numerical type is always allowed.
        User must be aware of precision loss and value change caused by range difference between two types.
        For example, a 64-bit float 3.1415926459 may be round to a 32-bit float 3.141592. Similarly, converting
        an integer 36 to Boolean may produce 1 because we truncate bits which can't be stored in the targeted type.


        Args:
            input: Input tensor to be cast.

            to: The data type to which the elements of the input tensor are cast.
                Strictly must be one of the types from DataType enum in TensorProto
        """

        schema = get_schema("Cast", 9, "")
        op = Op(self, "Cast", schema)
        return op(*self._prepare_inputs(schema, input), to=to)

    T_Compress = TypeVar(
        "T_Compress",
        BOOL,
        COMPLEX128,
        COMPLEX64,
        DOUBLE,
        FLOAT,
        FLOAT16,
        INT16,
        INT32,
        INT64,
        INT8,
        STRING,
        UINT16,
        UINT32,
        UINT64,
        UINT8,
    )

    T1_Compress: TypeAlias = BOOL

    def Compress(
        self, input: T_Compress, condition: T1_Compress, *, axis: Optional[int] = None
    ) -> T_Compress:
        r"""[🌐 Compress(9)](https://onnx.ai/onnx/operators/onnx__Compress.html#compress-9 "Online Documentation")


            Selects slices from an input tensor along a given axis where condition evaluates to True for each axis index.
            In case axis is not provided, input is flattened before elements are selected.
            Compress behaves like numpy.compress: https://docs.scipy.org/doc/numpy/reference/generated/numpy.compress.html


        Args:
            input: Tensor of rank r >= 1.

            condition: Rank 1 tensor of booleans to indicate which slices or data
                elements to be selected. Its length can be less than the input length
                alone the axis or the flattened input size if axis is not specified. In
                such cases data slices or elements exceeding the condition length are
                discarded.

            axis: (Optional) Axis along which to take slices. If not specified, input is
                flattened before elements being selected.
        """

        schema = get_schema("Compress", 9, "")
        op = Op(self, "Compress", schema)
        return op(*self._prepare_inputs(schema, input, condition), axis=axis)

    T_Constant: TypeAlias = Union[
        BOOL,
        COMPLEX128,
        COMPLEX64,
        DOUBLE,
        FLOAT,
        FLOAT16,
        INT16,
        INT32,
        INT64,
        INT8,
        STRING,
        UINT16,
        UINT32,
        UINT64,
        UINT8,
    ]

    def Constant(self, *, value: TensorProto) -> T_Constant:
        r"""[🌐 Constant(9)](https://onnx.ai/onnx/operators/onnx__Constant.html#constant-9 "Online Documentation")

        A constant tensor.

        Args:
            value: The value for the elements of the output tensor.
        """

        schema = get_schema("Constant", 9, "")
        op = Op(self, "Constant", schema)
        return op(value=value)

    T1_ConstantOfShape: TypeAlias = INT64

    T2_ConstantOfShape: TypeAlias = Union[
        BOOL,
        DOUBLE,
        FLOAT,
        FLOAT16,
        INT16,
        INT32,
        INT64,
        INT8,
        UINT16,
        UINT32,
        UINT64,
        UINT8,
    ]

    def ConstantOfShape(
        self, input: T1_ConstantOfShape, *, value: Optional[TensorProto] = None
    ) -> T2_ConstantOfShape:
        r"""[🌐 ConstantOfShape(9)](https://onnx.ai/onnx/operators/onnx__ConstantOfShape.html#constantofshape-9 "Online Documentation")


        Generate a tensor with given value and shape.


        Args:
            input: 1D tensor. The shape of the expected output tensor. If empty tensor
                is given, the output would be a scalar. All values must be >= 0.

            value: (Optional) The value of the output elements.Should be a one-element
                tensor. If not specified, it defaults to a tensor of value 0 and
                datatype float32
        """

        schema = get_schema("ConstantOfShape", 9, "")
        op = Op(self, "ConstantOfShape", schema)
        return op(*self._prepare_inputs(schema, input), value=value)

    T_Cosh = TypeVar("T_Cosh", DOUBLE, FLOAT, FLOAT16)

    def Cosh(self, input: T_Cosh) -> T_Cosh:
        r"""[🌐 Cosh(9)](https://onnx.ai/onnx/operators/onnx__Cosh.html#cosh-9 "Online Documentation")


        Calculates the hyperbolic cosine of the given input tensor element-wise.


        Args:
            input: (differentiable) Input tensor
        """

        schema = get_schema("Cosh", 9, "")
        op = Op(self, "Cosh", schema)
        return op(*self._prepare_inputs(schema, input))

    T_Erf = TypeVar(
        "T_Erf",
        DOUBLE,
        FLOAT,
        FLOAT16,
        INT16,
        INT32,
        INT64,
        INT8,
        UINT16,
        UINT32,
        UINT64,
        UINT8,
    )

    def Erf(self, input: T_Erf) -> T_Erf:
        r"""[🌐 Erf(9)](https://onnx.ai/onnx/operators/onnx__Erf.html#erf-9 "Online Documentation")


        Computes the error function of the given input tensor element-wise.


        Args:
            input: Input tensor
        """

        schema = get_schema("Erf", 9, "")
        op = Op(self, "Erf", schema)
        return op(*self._prepare_inputs(schema, input))

    T1_EyeLike = TypeVar(
        "T1_EyeLike",
        BOOL,
        DOUBLE,
        FLOAT,
        FLOAT16,
        INT16,
        INT32,
        INT64,
        INT8,
        UINT16,
        UINT32,
        UINT64,
        UINT8,
    )

    T2_EyeLike: TypeAlias = Union[
        BOOL,
        DOUBLE,
        FLOAT,
        FLOAT16,
        INT16,
        INT32,
        INT64,
        INT8,
        UINT16,
        UINT32,
        UINT64,
        UINT8,
    ]

    def EyeLike(
        self, input: T1_EyeLike, *, dtype: Optional[int] = None, k: int = 0
    ) -> T2_EyeLike:
        r"""[🌐 EyeLike(9)](https://onnx.ai/onnx/operators/onnx__EyeLike.html#eyelike-9 "Online Documentation")


        Generate a 2D tensor (matrix) with ones on the diagonal and zeros everywhere else. Only 2D
        tensors are supported, i.e. input T1 must be of rank 2. The shape of the output tensor is the
        same as the input tensor. The data type can be specified by the 'dtype' argument. If
        'dtype' is not specified, then the type of input tensor is used. By default, the main diagonal
        is populated with ones, but attribute 'k' can be used to populate upper or lower diagonals.
        The 'dtype' argument must be one of the data types specified in the 'DataType' enum field in the
        TensorProto message and be valid as an output type.


        Args:
            input: 2D input tensor to copy shape, and optionally, type information from.

            dtype: (Optional) The data type for the elements of the output tensor. If
                not specified,the data type of the input tensor T1 is used. If input
                tensor T1 is also notspecified, then type defaults to 'float'.

            k: (Optional) Index of the diagonal to be populated with ones. Default is 0.
                If T2 is the output, this op sets T2[i, i+k] = 1. k = 0 populates the
                main diagonal, k > 0 populates an upper diagonal,  and k < 0 populates a
                lower diagonal.
        """

        schema = get_schema("EyeLike", 9, "")
        op = Op(self, "EyeLike", schema)
        return op(*self._prepare_inputs(schema, input), dtype=dtype, k=k)

    T_Flatten = TypeVar(
        "T_Flatten",
        BOOL,
        COMPLEX128,
        COMPLEX64,
        DOUBLE,
        FLOAT,
        FLOAT16,
        INT16,
        INT32,
        INT64,
        INT8,
        STRING,
        UINT16,
        UINT32,
        UINT64,
        UINT8,
    )

    def Flatten(self, input: T_Flatten, *, axis: int = 1) -> T_Flatten:
        r"""[🌐 Flatten(9)](https://onnx.ai/onnx/operators/onnx__Flatten.html#flatten-9 "Online Documentation")


        Flattens the input tensor into a 2D matrix. If input tensor has shape
        (d_0, d_1, ... d_n) then the output will have shape
        (d_0 X d_1 ... d_(axis-1), d_axis X d_(axis+1) ... X dn).


        Args:
            input: A tensor of rank >= axis.

            axis: Indicate up to which input dimensions (exclusive) should be flattened
                to the outer dimension of the output. The value for axis must be in the
                range [0, R], where R is the rank of the input tensor. When axis = 0,
                the shape of the output tensor is (1, (d_0 X d_1 ... d_n), where the
                shape of the input tensor is (d_0, d_1, ... d_n).
        """

        schema = get_schema("Flatten", 9, "")
        op = Op(self, "Flatten", schema)
        return op(*self._prepare_inputs(schema, input), axis=axis)

    T_Gemm = TypeVar("T_Gemm", DOUBLE, FLOAT, FLOAT16, INT32, INT64, UINT32, UINT64)

    def Gemm(
        self,
        A: T_Gemm,
        B: T_Gemm,
        C: T_Gemm,
        *,
        alpha: float = 1.0,
        beta: float = 1.0,
        transA: int = 0,
        transB: int = 0,
    ) -> T_Gemm:
        r"""[🌐 Gemm(9)](https://onnx.ai/onnx/operators/onnx__Gemm.html#gemm-9 "Online Documentation")

        General Matrix multiplication:
        https://en.wikipedia.org/wiki/Basic_Linear_Algebra_Subprograms#Level_3

        A' = transpose(A) if transA else A

        B' = transpose(B) if transB else B

        Compute Y = alpha * A' * B' + beta * C, where input tensor A has shape (M, K) or (K, M),
        input tensor B has shape (K, N) or (N, K), input tensor C is broadcastable to shape (M, N),
        and output tensor Y has shape (M, N). A will be transposed before doing the
        computation if attribute transA is non-zero, same for B and transB.
        This operator supports **unidirectional broadcasting** (tensor C should be unidirectional broadcastable to tensor A * B); for more details please check `Broadcasting in ONNX <https://github.com/onnx/onnx/blob/master/docs/Broadcasting.md>`_.

        Args:
            A: Input tensor A. The shape of A should be (M, K) if transA is 0, or (K, M)
                if transA is non-zero.

            B: Input tensor B. The shape of B should be (K, N) if transB is 0, or (N, K)
                if transB is non-zero.

            C: Input tensor C. The shape of C should be unidirectional broadcastable to
                (M, N).

            alpha: Scalar multiplier for the product of input tensors A * B.

            beta: Scalar multiplier for input tensor C.

            transA: Whether A should be transposed

            transB: Whether B should be transposed
        """

        schema = get_schema("Gemm", 9, "")
        op = Op(self, "Gemm", schema)
        return op(
            *self._prepare_inputs(schema, A, B, C),
            alpha=alpha,
            beta=beta,
            transA=transA,
            transB=transB,
        )

    T_Greater = TypeVar(
        "T_Greater",
        DOUBLE,
        FLOAT,
        FLOAT16,
        INT16,
        INT32,
        INT64,
        INT8,
        UINT16,
        UINT32,
        UINT64,
        UINT8,
    )

    T1_Greater: TypeAlias = BOOL

    def Greater(self, A: T_Greater, B: T_Greater) -> T1_Greater:
        r"""[🌐 Greater(9)](https://onnx.ai/onnx/operators/onnx__Greater.html#greater-9 "Online Documentation")


        Returns the tensor resulted from performing the `greater` logical operation
        elementwise on the input tensors `A` and `B` (with Numpy-style broadcasting support).

        This operator supports **multidirectional (i.e., Numpy-style) broadcasting**; for more details please check `Broadcasting in ONNX <https://github.com/onnx/onnx/blob/master/docs/Broadcasting.md>`_.


        Args:
            A: First input operand for the logical operator.

            B: Second input operand for the logical operator.
        """

        schema = get_schema("Greater", 9, "")
        op = Op(self, "Greater", schema)
        return op(*self._prepare_inputs(schema, A, B))

    T1_IsNaN = TypeVar("T1_IsNaN", DOUBLE, FLOAT, FLOAT16)

    T2_IsNaN: TypeAlias = BOOL

    def IsNaN(self, X: T1_IsNaN) -> T2_IsNaN:
        r"""[🌐 IsNaN(9)](https://onnx.ai/onnx/operators/onnx__IsNaN.html#isnan-9 "Online Documentation")

        Returns which elements of the input are NaN.

        Args:
            X: input
        """

        schema = get_schema("IsNaN", 9, "")
        op = Op(self, "IsNaN", schema)
        return op(*self._prepare_inputs(schema, X))

    T_Less = TypeVar(
        "T_Less",
        DOUBLE,
        FLOAT,
        FLOAT16,
        INT16,
        INT32,
        INT64,
        INT8,
        UINT16,
        UINT32,
        UINT64,
        UINT8,
    )

    T1_Less: TypeAlias = BOOL

    def Less(self, A: T_Less, B: T_Less) -> T1_Less:
        r"""[🌐 Less(9)](https://onnx.ai/onnx/operators/onnx__Less.html#less-9 "Online Documentation")


        Returns the tensor resulted from performing the `less` logical operation
        elementwise on the input tensors `A` and `B` (with Numpy-style broadcasting support).

        This operator supports **multidirectional (i.e., Numpy-style) broadcasting**; for more details please check `Broadcasting in ONNX <https://github.com/onnx/onnx/blob/master/docs/Broadcasting.md>`_.


        Args:
            A: First input operand for the logical operator.

            B: Second input operand for the logical operator.
        """

        schema = get_schema("Less", 9, "")
        op = Op(self, "Less", schema)
        return op(*self._prepare_inputs(schema, A, B))

    T_MatMul = TypeVar("T_MatMul", DOUBLE, FLOAT, FLOAT16, INT32, INT64, UINT32, UINT64)

    def MatMul(self, A: T_MatMul, B: T_MatMul) -> T_MatMul:
        r"""[🌐 MatMul(9)](https://onnx.ai/onnx/operators/onnx__MatMul.html#matmul-9 "Online Documentation")


        Matrix product that behaves like [numpy.matmul](https://numpy.org/doc/stable/reference/generated/numpy.matmul.html).


        Args:
            A: N-dimensional matrix A

            B: N-dimensional matrix B
        """

        schema = get_schema("MatMul", 9, "")
        op = Op(self, "MatMul", schema)
        return op(*self._prepare_inputs(schema, A, B))

    T1_MaxUnpool = TypeVar("T1_MaxUnpool", DOUBLE, FLOAT, FLOAT16)

    T2_MaxUnpool: TypeAlias = INT64

    def MaxUnpool(
        self,
        X: T1_MaxUnpool,
        I: T2_MaxUnpool,
        output_shape: Optional[T2_MaxUnpool] = None,
        *,
        kernel_shape: Sequence[int],
        pads: Optional[Sequence[int]] = None,
        strides: Optional[Sequence[int]] = None,
    ) -> T1_MaxUnpool:
        r"""[🌐 MaxUnpool(9)](https://onnx.ai/onnx/operators/onnx__MaxUnpool.html#maxunpool-9 "Online Documentation")


        MaxUnpool essentially computes the partial inverse of the MaxPool op.
         The input information to this op is typically the output information from a MaxPool op. The first
         input tensor X is the tensor that needs to be unpooled, which is typically the pooled tensor (first output)
         from MaxPool. The second input tensor, I, contains the indices to the (locally maximal) elements corresponding
         to the elements in the first input tensor X. Input tensor I is typically the second output of the MaxPool op.
         The third (optional) input is a tensor that specifies the output size of the unpooling operation.

        MaxUnpool is intended to do 'partial' inverse of the MaxPool op. 'Partial' because all the non-maximal
         values from the original input to MaxPool are set to zero in the output of the MaxUnpool op. Pooling
         the result of an unpooling operation should give back the original input to the unpooling op.

        MaxUnpool can produce the same output size for several input sizes, which makes unpooling op ambiguous.
         The third input argument, output_size, is meant to disambiguate the op and produce output tensor of
         known/predictable size.

        In addition to the inputs, MaxUnpool takes three attributes, namely kernel_shape, strides, and pads,
         which define the exact unpooling op. The attributes typically have the same values as the corresponding
         pooling op that the unpooling op is trying to invert.


        Args:
            X: Input data tensor that has to be unpooled. This tensor is typically the
                first output of the MaxPool op.Dimensions for image case are (N x C x H
                x W), where N is the batch size, C is the number of channels, and H and
                W are the height and the width of the data. For non-image case, the
                dimensions are in the form of (N x C x D1 x D2 ... Dn), where N is the
                batch size. Optionally, if dimension denotation is in effect, the
                operation expects the input data tensor to arrive with the dimension
                denotation of [DATA_BATCH, DATA_CHANNEL, DATA_FEATURE, DATA_FEATURE
                ...].

            I: Input data tensor containing the indices corresponding to elements in the
                first input tensor X.This tensor is typically the second output of the
                MaxPool op.Dimensions must be the same as input tensor X. The indices
                are linear, i.e. computed considering the tensor as flattened 1-D
                tensor, assuming row-major storage. Also, the linear indices should not
                consider padding. So the values in indices are in the range [0, N x C x
                D1 x ... x Dn).

            output_shape: (optional) The shape of the output can be explicitly set which
                will cause pads values to be auto generated. If 'output_shape' is
                specified, 'pads' values are ignored.

            kernel_shape: The size of the kernel along each axis.

            pads: Padding for the beginning and ending along each spatial axis, it can
                take any value greater than or equal to 0. The value represent the
                number of pixels added to the beginning and end part of the
                corresponding axis. `pads` format should be as follow [x1_begin,
                x2_begin...x1_end, x2_end,...], where xi_begin the number of pixels
                added at the beginning of axis `i` and xi_end, the number of pixels
                added at the end of axis `i`. This attribute cannot be used
                simultaneously with auto_pad attribute. If not present, the padding
                defaults to 0 along start and end of each spatial axis.

            strides: Stride along each spatial axis.
        """

        schema = get_schema("MaxUnpool", 9, "")
        op = Op(self, "MaxUnpool", schema)
        return op(
            *self._prepare_inputs(schema, X, I, output_shape),
            kernel_shape=kernel_shape,
            pads=pads,
            strides=strides,
        )

    T_MeanVarianceNormalization = TypeVar(
        "T_MeanVarianceNormalization", DOUBLE, FLOAT, FLOAT16
    )

    def MeanVarianceNormalization(
        self, X: T_MeanVarianceNormalization, *, axes: Sequence[int] = (0, 2, 3)
    ) -> T_MeanVarianceNormalization:
        r"""[🌐 MeanVarianceNormalization(9)](https://onnx.ai/onnx/operators/onnx__MeanVarianceNormalization.html#meanvariancenormalization-9 "Online Documentation")


              A MeanVarianceNormalization Function: Perform mean variance normalization
              on the input tensor X using formula: <br/> ``` (X-EX)/sqrt(E(X-EX)^2) ```


        Args:
            X: Input tensor

            axes: A list of integers, along which to reduce. The default is to calculate
                along axes [0,2,3] for calculating mean and variance along each channel.
                Two variables with the same C-coordinate are associated with the same
                mean and variance.
        """

        schema = get_schema("MeanVarianceNormalization", 9, "")
        op = Op(self, "MeanVarianceNormalization", schema)
        return op(*self._prepare_inputs(schema, X), axes=axes)

    T_NonZero = TypeVar(
        "T_NonZero",
        BOOL,
        COMPLEX128,
        COMPLEX64,
        DOUBLE,
        FLOAT,
        FLOAT16,
        INT16,
        INT32,
        INT64,
        INT8,
        STRING,
        UINT16,
        UINT32,
        UINT64,
        UINT8,
    )

    def NonZero(self, X: T_NonZero) -> INT64:
        r"""[🌐 NonZero(9)](https://onnx.ai/onnx/operators/onnx__NonZero.html#nonzero-9 "Online Documentation")


            Returns the indices of the elements that are non-zero
            (in row-major order - by dimension).
            NonZero behaves similar to numpy.nonzero:
            https://docs.scipy.org/doc/numpy/reference/generated/numpy.nonzero.html,
            but for scalar input, NonZero produces output shape (0, N) instead of (1, N), which is different from Numpy's behavior.


        Args:
            X: input
        """

        schema = get_schema("NonZero", 9, "")
        op = Op(self, "NonZero", schema)
        return op(*self._prepare_inputs(schema, X))

    T1_OneHot = TypeVar(
        "T1_OneHot",
        DOUBLE,
        FLOAT,
        FLOAT16,
        INT16,
        INT32,
        INT64,
        INT8,
        UINT16,
        UINT32,
        UINT64,
        UINT8,
    )

    T2_OneHot = TypeVar(
        "T2_OneHot",
        DOUBLE,
        FLOAT,
        FLOAT16,
        INT16,
        INT32,
        INT64,
        INT8,
        UINT16,
        UINT32,
        UINT64,
        UINT8,
    )

    T3_OneHot = TypeVar(
        "T3_OneHot",
        BOOL,
        COMPLEX128,
        COMPLEX64,
        DOUBLE,
        FLOAT,
        FLOAT16,
        INT16,
        INT32,
        INT64,
        INT8,
        STRING,
        UINT16,
        UINT32,
        UINT64,
        UINT8,
    )

    def OneHot(
        self, indices: T1_OneHot, depth: T2_OneHot, values: T3_OneHot, *, axis: int = -1
    ) -> T3_OneHot:
        r"""[🌐 OneHot(9)](https://onnx.ai/onnx/operators/onnx__OneHot.html#onehot-9 "Online Documentation")


            Produces a one-hot tensor based on inputs.
            The locations represented by the index values in the 'indices' input tensor will have 'on_value'
            and the other locations will have 'off_value' in the output tensor, where 'on_value' and 'off_value'
            are specified as part of required input argument 'values', which is a two-element tensor of format
            [off_value, on_value]. The rank of the output tensor will be one greater than the rank of the
            input tensor. The additional dimension is for one-hot representation. The additional dimension will
            be inserted at the position specified by 'axis'. If 'axis' is not specified then then additional
            dimension will be inserted as the innermost dimension, i.e. axis=-1. The size of the additional
            dimension is specified by required scalar input 'depth'. The type of the output tensor is the same
            as the type of the 'values' input. Any entries in the 'indices' input tensor with values outside
            the range [0, depth) will result in one-hot representation with all 'off_value' values in the
            output tensor.


        Args:
            indices: Input tensor containing indices. The values must be non-negative
                integers. Any entries in the 'indices' input tensor with values outside
                the range [0, depth) will result in one-hot representation with all
                'off_value' values in the output tensor.In case 'indices' is of
                non-integer type, the values will be casted to int64 before use.

            depth: Scalar or rank 1 tensor containing exactly one element, specifying
                the number of classes in one-hot tensor. This is also the size of the
                one-hot dimension (specified by 'axis' attribute) added on in the output
                tensor. The values in the 'indices' input tensor are expected to be in
                the range [0, depth). In case 'depth' is of non-integer type, it will be
                casted to int64 before use.

            values: Rank 1 tensor containing exactly two elements, in the format
                [off_value, on_value], where 'on_value' is the value used for filling
                locations specified in 'indices' input tensor, and 'off_value' is the
                value used for filling locations other than those specified in 'indices'
                input tensor.

            axis: (Optional) Axis along which one-hot representation in added. Default:
                axis=-1. axis=-1 means that the additional dimension will be inserted as
                the innermost/last dimension in the output tensor.
        """

        schema = get_schema("OneHot", 9, "")
        op = Op(self, "OneHot", schema)
        return op(*self._prepare_inputs(schema, indices, depth, values), axis=axis)

    T_PRelu = TypeVar("T_PRelu", DOUBLE, FLOAT, FLOAT16, INT32, INT64, UINT32, UINT64)

    def PRelu(self, X: T_PRelu, slope: T_PRelu) -> T_PRelu:
        r"""[🌐 PRelu(9)](https://onnx.ai/onnx/operators/onnx__PRelu.html#prelu-9 "Online Documentation")


        PRelu takes input data (Tensor<T>) and slope tensor as input, and produces one
        output data (Tensor<T>) where the function `f(x) = slope * x for x < 0`,
        `f(x) = x for x >= 0`., is applied to the data tensor elementwise.
        This operator supports **unidirectional broadcasting** (tensor slope should be unidirectional broadcastable to input tensor X); for more details please check `Broadcasting in ONNX <https://github.com/onnx/onnx/blob/master/docs/Broadcasting.md>`_.

        Args:
            X: (differentiable) Input tensor

            slope: (differentiable) Slope tensor. The shape of slope can be smaller than
                first input X; if so, its shape must be unidirectional broadcastable to
                X
        """

        schema = get_schema("PRelu", 9, "")
        op = Op(self, "PRelu", schema)
        return op(*self._prepare_inputs(schema, X, slope))

    V_Scan = TypeVar(
        "V_Scan",
        BOOL,
        COMPLEX128,
        COMPLEX64,
        DOUBLE,
        FLOAT,
        FLOAT16,
        INT16,
        INT32,
        INT64,
        INT8,
        STRING,
        UINT16,
        UINT32,
        UINT64,
        UINT8,
    )

    def Scan(
        self,
        *initial_state_and_scan_inputs: V_Scan,
        body: GraphProto,
        num_scan_inputs: int,
        scan_input_axes: Optional[Sequence[int]] = None,
        scan_input_directions: Optional[Sequence[int]] = None,
        scan_output_axes: Optional[Sequence[int]] = None,
        scan_output_directions: Optional[Sequence[int]] = None,
    ) -> V_Scan:
        r"""[🌐 Scan(9)](https://onnx.ai/onnx/operators/onnx__Scan.html#scan-9 "Online Documentation")


        Scan can be used to iterate over one or more scan_input tensors,
        constructing zero or more scan_output tensors. It combines ideas from general recurrences,
        functional programming constructs such as scan, fold, map, and zip, and is intended to enable
        generalizations of RNN-like constructs for sequence-to-sequence processing.
        Other tensors (referred to as state_variables here) can be used to carry a state
        when iterating from one element to another (similar to hidden-state in RNNs, also referred
        to as loop-carried dependences in the context of loops).
        Many common usages involve a single scan_input tensor (where functionality
        similar to scan, fold and map can be obtained). When more than one scan_input is used,
        a behavior similar to zip is obtained.

        The attribute body must be a graph, specifying the computation to be performed in
        every iteration. It takes as input the current values of the state_variables and
        the current iterated element of the scan_inputs. It must return the (updated) values
        of the state_variables and zero or more scan_output_element tensors. The values of the
        scan_output_element tensors are concatenated over all the iterations to produce the
        scan_output values of the scan construct (similar to the concatenated intermediate
        hidden-state values of RNN-like constructs). All the output tensors (state_variables as
        well as scan_output_element tensors) are required to have the same shape in each iteration
        of the loop (a restriction imposed to enable efficient memory allocation).

        Note that the iterated element passed to the body subgraph does not have a sequence
        axis. It will have a rank one less than the rank of the corresponding scan_input.

        The scan operation returns the final values of the state_variables as well as the
        scan_outputs.

        The optional attribute scan_input_directions specifies the direction (forward or backward)
        for each scan input. If this attribute is omitted, all sequences are scanned in the forward
        direction. A bidirectional scan may be performed by specifying the same tensor input twice
        in the scan_inputs, once with a forward direction, and once with a backward direction.

        The scan_output of the operation is produced by concatenating the scan_output_element
        values produced by the body in each iteration.  The optional attribute scan_output_directions
        specifies the direction in which scan_output is constructed (by appending or prepending the
        scan_output_element to scan_output in each iteration) for each scan_output. If this attribute
        is omitted, the scan_output_element is appended to the scan_output in each iteration.

        The optional attribute scan_input_axes specifies the axis to be scanned for each scan_input.
        If omitted, every scan_input will be scanned in axis 0. For example, if axis 0 is the
        batch axis and axis 1 is the time axis (to be scanned), specify an axis value of 1.
        Note that scanning a non-zero axis may be less efficient than scanning axis zero.

        The optional attribute scan_output_axes specifies the axis along which the scan_outputs
        are accumulated for each scan_output. For example, if axis 1 is the time axis (to be
        scanned) for both inputs and outputs, specify a scan_input axis and scan_output axis
        value of 1.

        Note that because of the ONNX restriction that only the last parameter of an operator can
        be variadic, the initial-states and scan-inputs are listed together as one input parameter.
        Similarly, the final-states and scan-outputs are listed together as one output parameter.
        The attribute num_scan_inputs indicates the number M of scan-inputs.

        The behavior of

            Scan <
                num_scan_inputs = m,
                body = loop-body,
                scan_input_axes = [axis_1, ..., axis_m]
            > (init_1, ..., init_n, scan_1, ..., scan_m)

        is equivalent to the following pseudo-code:

            // scan_i.shape[axis_i] denotes the (max) sequence-length of scan_i
            // scan_i.shape[axis_i] is required to be equal to scan_j.shape[axis_j] for all i,j.
            sequence_length = scan_1.shape[axis_1];

            // initialize state-variables
            st_1 = init_1; ... st_n = init_n;
            // initialize scan-output variables: [] denotes an empty tensor
            scan_out_1 = []; ...; scan_out_k = [];
            // identify number of iterations:

            // execute loop
            for (int t = 0; t < sequence_length; ++t) {
                // generate the scan-input elements: the notation T<axis=k>[t] indicates the sub-tensor
                // of rank one less than T obtained by indexing T at position t along axis k.
                si_1 = scan_1<axis=axis_1>[t];
                ... ;
                si_m = scan_m<axis=axis_m>[t];
                // execute loop-body
                st_1, ..., st_n, so_1, ..., so_k = loop-body(st_1, ..., st_n, si_1, ..., si_m)
                // accumulate the scan-output elements
                scan_out_1 = Concat<axis=0>(scan_out_1, so_1); ... ; scan_out_k = Concat<axis=0>(scan_out_k, so_k);
            }

            return st_1, ..., st_n, scan_out_1, ..., scan_out_k;

        *Sample usage: Encoding RNN using a Scan*

        The following example shows how a simple RNN over an input tensor %X, with weight tensor %Wi,
        recurrence weight tensor %Ri, bias tensors %Wbi and %Rbi, and initial hidden-state %H_0 can
        be encoded as a ScanLoop. Note that the loop-body is a nested graph, and it directly computes
        %Wi, %Ri, %Wbi, and %Rbi (typically constants or initializers in the body graph). If these
        values are computed in the outer graph, they need to be passed in as extra state_variables.

            graph rnn-encoding {
              %H_0 = ...
              %X = ...
              %Y_h, %Y = Scan[body = <graph rnn-cell-1>, num_scan_inputs=1](%H_0, %X)
              return %Y, %Y_h
            }

            graph rnn-cell-1 (
              %H_tminus1[FLOAT, tensor]
              %X_t[FLOAT, tensor]
            ) {
              %Wi = ...
              %Ri = ...
              %Wbi = ...
              %Rbi = ...
              %t1 = X_t * (Wi^T)
              %t2 = H_tminus1*(Ri^T)
              %t3 = Add(%t1, %t2)
              %t4 = Add(%t3, %Wbi)
              %t5 = Add(%t4, %Rbi)
              %Ht = Tanh(%t5)
              %Accumulate = Identity(%Ht)
              return %Ht, %Accumulate
            }



        Args:
            initial_state_and_scan_inputs: (variadic, heterogeneous) Initial values of
                the loop's N state variables followed by M scan_inputs

            body: The graph run each iteration. It has N+M inputs: (loop state
                variables..., scan_input_elts...). It has N+K outputs: (loop state
                variables..., scan_output_elts...). Each scan_output is created by
                concatenating the value of the specified scan_output_elt value at the
                end of each iteration of the loop. It is an error if the dimensions of
                these values change across loop iterations.

            num_scan_inputs: An attribute specifying the number of scan_inputs M.

            scan_input_axes: An optional list of M flags. The i-th element of the list
                specifies the axis to be scanned (the sequence axis) for the i-th
                scan_input. If omitted, 0 will be used as the scan axis for every
                scan_input.

            scan_input_directions: An optional list of M flags. The i-th element of the
                list specifies the direction to be scanned for the i-th scan_input
                tensor: 0 indicates forward direction and 1 indicates reverse direction.
                If omitted, all scan_input tensors will be scanned in the forward
                direction.

            scan_output_axes: An optional list of K flags. The i-th element of the list
                specifies the axis for the i-th scan_output. The scan outputs are
                accumulated along the specified axis. If omitted, 0 will be used as the
                scan axis for every scan_output.

            scan_output_directions: An optional list of K flags, one for each
                scan_output. The i-th element of the list specifies whether the i-th
                scan_output should be constructed by appending or prepending a new value
                in each iteration: 0 indicates appending and 1 indicates prepending. If
                omitted, all scan_output tensors will be produced by appending a value
                in each iteration.
        """

        schema = get_schema("Scan", 9, "")
        op = Op(self, "Scan", schema)
        return op(
            *self._prepare_inputs(schema, *initial_state_and_scan_inputs),
            body=body,
            num_scan_inputs=num_scan_inputs,
            scan_input_axes=scan_input_axes,
            scan_input_directions=scan_input_directions,
            scan_output_axes=scan_output_axes,
            scan_output_directions=scan_output_directions,
        )

    T_Scatter = TypeVar(
        "T_Scatter",
        BOOL,
        COMPLEX128,
        COMPLEX64,
        DOUBLE,
        FLOAT,
        FLOAT16,
        INT16,
        INT32,
        INT64,
        INT8,
        STRING,
        UINT16,
        UINT32,
        UINT64,
        UINT8,
    )

    Tind_Scatter = TypeVar("Tind_Scatter", INT32, INT64)

    def Scatter(
        self,
        data: T_Scatter,
        indices: Tind_Scatter,
        updates: T_Scatter,
        *,
        axis: int = 0,
    ) -> T_Scatter:
        r"""[🌐 Scatter(9)](https://onnx.ai/onnx/operators/onnx__Scatter.html#scatter-9 "Online Documentation")


        Given `data`, `updates` and `indices` input tensors of rank r >= 1, write the values provided by `updates`
        into the first input, `data`, along `axis` dimension of `data` (by default outer-most one as axis=0) at corresponding `indices`.
        For each entry in `updates`, the target index in `data` is specified by corresponding entry in `indices`
        for dimension = axis, and index in source for dimension != axis. For instance, in a 2-D tensor case,
        data[indices[i][j]][j] = updates[i][j] if axis = 0, or data[i][indices[i][j]] = updates[i][j] if axis = 1,
        where i and j are loop counters from 0 up to the respective size in `updates` - 1.
        Example 1:
          data = [
              [0.0, 0.0, 0.0],
              [0.0, 0.0, 0.0],
              [0.0, 0.0, 0.0],
          ]
          indices = [
              [1, 0, 2],
              [0, 2, 1],
          ]
          updates = [
              [1.0, 1.1, 1.2],
              [2.0, 2.1, 2.2],
          ]
          output = [
              [2.0, 1.1, 0.0]
              [1.0, 0.0, 2.2]
              [0.0, 2.1, 1.2]
          ]
        Example 2:
          data = [[1.0, 2.0, 3.0, 4.0, 5.0]]
          indices = [[1, 3]]
          updates = [[1.1, 2.1]]
          axis = 1
          output = [[1.0, 1.1, 3.0, 2.1, 5.0]]


        Args:
            data: Tensor of rank r >= 1.

            indices: Tensor of int32/int64 indices, of r >= 1 (same rank as input).

            updates: Tensor of rank r >=1 (same rank and shape as indices)

            axis: Which axis to scatter on. Negative value means counting dimensions
                from the back. Accepted range is [-r, r-1]
        """

        schema = get_schema("Scatter", 9, "")
        op = Op(self, "Scatter", schema)
        return op(*self._prepare_inputs(schema, data, indices, updates), axis=axis)

    T_Shrink = TypeVar(
        "T_Shrink",
        DOUBLE,
        FLOAT,
        FLOAT16,
        INT16,
        INT32,
        INT64,
        INT8,
        UINT16,
        UINT32,
        UINT64,
        UINT8,
    )

    def Shrink(self, input: T_Shrink, *, bias: float = 0.0, lambd: float = 0.5) -> T_Shrink:
        r"""[🌐 Shrink(9)](https://onnx.ai/onnx/operators/onnx__Shrink.html#shrink-9 "Online Documentation")


        Shrink takes one input data (Tensor<numeric>) and produces one Tensor output,
        having same datatype and shape with input. It has two attributes, lambd and
        bias. The formula of this operator is: If x < -lambd, y = x + bias;
        If x > lambd, y = x - bias; Otherwise, y = 0.


        Args:
            input: (differentiable) The input data as Tensor.

            bias: The bias value added to output. Default is 0.

            lambd: The lambd value for the Shrink formulation. Default is 0.5.
        """

        schema = get_schema("Shrink", 9, "")
        op = Op(self, "Shrink", schema)
        return op(*self._prepare_inputs(schema, input), bias=bias, lambd=lambd)

    T_Sign = TypeVar(
        "T_Sign",
        DOUBLE,
        FLOAT,
        FLOAT16,
        INT16,
        INT32,
        INT64,
        INT8,
        UINT16,
        UINT32,
        UINT64,
        UINT8,
    )

    def Sign(self, input: T_Sign) -> T_Sign:
        r"""[🌐 Sign(9)](https://onnx.ai/onnx/operators/onnx__Sign.html#sign-9 "Online Documentation")


        Calculate the sign of the given input tensor element-wise.
        If input > 0, output 1. if input < 0, output -1. if input == 0, output 0.


        Args:
            input: Input tensor
        """

        schema = get_schema("Sign", 9, "")
        op = Op(self, "Sign", schema)
        return op(*self._prepare_inputs(schema, input))

    T_Sinh = TypeVar("T_Sinh", DOUBLE, FLOAT, FLOAT16)

    def Sinh(self, input: T_Sinh) -> T_Sinh:
        r"""[🌐 Sinh(9)](https://onnx.ai/onnx/operators/onnx__Sinh.html#sinh-9 "Online Documentation")


        Calculates the hyperbolic sine of the given input tensor element-wise.


        Args:
            input: (differentiable) Input tensor
        """

        schema = get_schema("Sinh", 9, "")
        op = Op(self, "Sinh", schema)
        return op(*self._prepare_inputs(schema, input))

    T_TfIdfVectorizer = TypeVar("T_TfIdfVectorizer", INT32, INT64, STRING)

    T1_TfIdfVectorizer: TypeAlias = FLOAT

    def TfIdfVectorizer(
        self,
        X: T_TfIdfVectorizer,
        *,
        max_gram_length: int,
        max_skip_count: int,
        min_gram_length: int,
        mode: str,
        ngram_counts: Sequence[int],
        ngram_indexes: Sequence[int],
        pool_int64s: Optional[Sequence[int]] = None,
        pool_strings: Optional[Sequence[str]] = None,
        weights: Optional[Sequence[float]] = None,
    ) -> T1_TfIdfVectorizer:
        r"""[🌐 TfIdfVectorizer(9)](https://onnx.ai/onnx/operators/onnx__TfIdfVectorizer.html#tfidfvectorizer-9 "Online Documentation")


        This transform extracts n-grams from the input sequence and save them as a vector. Input can
        be either a 1-D or 2-D tensor. For 1-D input, output is the n-gram representation of that input.
        For 2-D input, the output is also a  2-D tensor whose i-th row is the n-gram representation of the i-th input row.
        More specifically, if input shape is [C], the corresponding output shape would be [max(ngram_indexes) + 1].
        If input shape is [N, C], this operator produces a [N, max(ngram_indexes) + 1]-tensor.

        In contrast to standard n-gram extraction, here, the indexes of extracting an n-gram from the original
        sequence are not necessarily consecutive numbers. The discontinuity between indexes are controlled by the number of skips.
        If the number of skips is 2, we should skip two tokens when scanning through the original sequence.
        Let's consider an example. Assume that input sequence is [94, 17, 36, 12, 28] and the number of skips is 2.
        The associated 2-grams are [94, 12] and [17, 28] respectively indexed by [0, 3] and [1, 4].
        If the number of skips becomes 0, the 2-grams generated are [94, 17], [17, 36], [36, 12], [12, 28]
        indexed by [0, 1], [1, 2], [2, 3], [3, 4], respectively.

        The output vector (denoted by Y) stores the count of each n-gram;
        Y[ngram_indexes[i]] indicates the times that the i-th n-gram is found. The attribute ngram_indexes is used to determine the mapping
        between index i and the corresponding n-gram's output coordinate. If pool_int64s is [94, 17, 17, 36], ngram_indexes is [1, 0],
        ngram_counts=[0, 0], then the Y[0] (first element in Y) and Y[1] (second element in Y) are the counts of [17, 36] and [94, 17],
        respectively. An n-gram which cannot be found in pool_strings/pool_int64s should be ignored and has no effect on the output.
        Note that we may consider all skips up to S when generating the n-grams.

        The examples used above are true if mode is "TF". If mode is "IDF", all the counts larger than 1 would be truncated to 1 and
        the i-th element in weights would be used to scale (by multiplication) the count of the i-th n-gram in pool. If mode is "TFIDF",
        this operator first computes the counts of all n-grams and then scale them by the associated values in the weights attribute.

        Only one of pool_strings and pool_int64s can be set. If pool_int64s is set, the input should be an integer tensor.
        If pool_strings is set, the input must be a string tensor.


        Args:
            X: (non-differentiable) Input for n-gram extraction

            max_gram_length: Maximum n-gram length. If this value is 3, 3-grams will be
                used to generate the output.

            max_skip_count: Maximum number of items (integers/strings) to be skipped
                when constructing an n-gram from X. If max_skip_count=1,
                min_gram_length=2, max_gram_length=3, this operator may generate 2-grams
                with skip_count=0 and skip_count=1, and 3-grams with skip_count=0 and
                skip_count=1

            min_gram_length: Minimum n-gram length. If this value is 2 and
                max_gram_length is 3, output may contain counts of 2-grams and 3-grams.

            mode: The weighting criteria. It can be one of "TF" (term frequency), "IDF"
                (inverse document frequency), and "TFIDF" (the combination of TF and
                IDF)

            ngram_counts: The starting indexes of 1-grams, 2-grams, and so on in pool.
                It is useful when determining the boundary between two consecutive
                collections of n-grams. For example, if ngram_counts is [0, 17, 36], the
                first index (zero-based) of 1-gram/2-gram/3-gram in pool are 0/17/36.
                This format is essentially identical to CSR (or CSC) sparse matrix
                format, and we choose to use this due to its popularity.

            ngram_indexes: list of int64s (type: AttributeProto::INTS). This list is
                parallel to the specified 'pool_*' attribute. The i-th element in
                ngram_indexes indicate the coordinate of the i-th n-gram in the output
                tensor.

            pool_int64s: List of int64 n-grams learned from the training set. Either
                this or pool_strings attributes must be present but not both. It's an
                1-D tensor starting with the collections of all 1-grams and ending with
                the collections of n-grams. The i-th element in pool stores the n-gram
                that should be mapped to coordinate ngram_indexes[i] in the output
                vector.

            pool_strings: List of strings n-grams learned from the training set. Either
                this or pool_int64s attributes must be present but not both. It's an 1-D
                tensor starting with the collections of all 1-grams and ending with the
                collections of n-grams. The i-th element in pool stores the n-gram that
                should be mapped to coordinate ngram_indexes[i] in the output vector.

            weights: list of floats. This attribute stores the weight of each n-gram in
                pool. The i-th element in weights is the weight of the i-th n-gram in
                pool. Its length equals to the size of ngram_indexes. By default,
                weights is an all-one tensor.This attribute is used when mode is "IDF"
                or "TFIDF" to scale the associated word counts.
        """

        schema = get_schema("TfIdfVectorizer", 9, "")
        op = Op(self, "TfIdfVectorizer", schema)
        return op(
            *self._prepare_inputs(schema, X),
            max_gram_length=max_gram_length,
            max_skip_count=max_skip_count,
            min_gram_length=min_gram_length,
            mode=mode,
            ngram_counts=ngram_counts,
            ngram_indexes=ngram_indexes,
            pool_int64s=pool_int64s,
            pool_strings=pool_strings,
            weights=weights,
        )

    T_Upsample = TypeVar(
        "T_Upsample",
        BOOL,
        COMPLEX128,
        COMPLEX64,
        DOUBLE,
        FLOAT,
        FLOAT16,
        INT16,
        INT32,
        INT64,
        INT8,
        STRING,
        UINT16,
        UINT32,
        UINT64,
        UINT8,
    )

    def Upsample(self, X: T_Upsample, scales: FLOAT, *, mode: str = "nearest") -> T_Upsample:
        r"""[🌐 Upsample(9)](https://onnx.ai/onnx/operators/onnx__Upsample.html#upsample-9 "Online Documentation")


        Upsample the input tensor.
        Each dimension value of the output tensor is:
          output_dimension = floor(input_dimension * scale).


        Args:
            X: N-D tensor

            scales: The scale array along each dimension. It takes value greater than or
                equal to 1. The number of elements of 'scales' should be the same as the
                rank of input 'X'.

            mode: Two interpolation modes: nearest (default), and linear (including
                bilinear, trilinear, etc)
        """

        schema = get_schema("Upsample", 9, "")
        op = Op(self, "Upsample", schema)
        return op(*self._prepare_inputs(schema, X, scales), mode=mode)

    B_Where: TypeAlias = BOOL

    T_Where = TypeVar(
        "T_Where",
        BOOL,
        COMPLEX128,
        COMPLEX64,
        DOUBLE,
        FLOAT,
        FLOAT16,
        INT16,
        INT32,
        INT64,
        INT8,
        STRING,
        UINT16,
        UINT32,
        UINT64,
        UINT8,
    )

    def Where(self, condition: B_Where, X: T_Where, Y: T_Where) -> T_Where:
        r"""[🌐 Where(9)](https://onnx.ai/onnx/operators/onnx__Where.html#where-9 "Online Documentation")


        Return elements, either from X or Y, depending on condition.
        Where behaves like
        [numpy.where](https://docs.scipy.org/doc/numpy/reference/generated/numpy.where.html)
        with three parameters.

        This operator supports **multidirectional (i.e., Numpy-style) broadcasting**; for more details please check `Broadcasting in ONNX <https://github.com/onnx/onnx/blob/master/docs/Broadcasting.md>`_.

        Args:
            condition: (non-differentiable) When True (nonzero), yield X, otherwise
                yield Y

            X: (differentiable) values selected at indices where condition is True

            Y: (differentiable) values selected at indices where condition is False
        """

        schema = get_schema("Where", 9, "")
        op = Op(self, "Where", schema)
        return op(*self._prepare_inputs(schema, condition, X, Y))



# Copyright (c) Microsoft Corporation.
# Licensed under the MIT License.
import unittest
from typing import Sequence

import numpy as np
import onnx
from onnx_ir.passes.common import onnx_checker, shape_inference
from parameterized import parameterized

from onnxscript import ir
from onnxscript.rewriter import MatchingTracer, MatchStatus, testing
from onnxscript.rewriter.rules.common import _matmul_add_to_gemm


class _MatMulAddToGemmTestBase(unittest.TestCase):
    @property
    def rng(self):
        return np.random.default_rng(20250607)

    def clone_model(self, model: ir.Model) -> ir.Model:
        return ir.from_proto(ir.to_proto(model))

    def get_test_model(
        self,
        input_shape: ir.Shape,
        weight_shape: ir.Shape,
        transA: bool = False,
        transB: bool = False,
        permA: Sequence[int] = [1, 0],
        permB: Sequence[int] = [1, 0],
        weight_as_inputs: bool = False,
        bias_as_inputs: bool = False,
    ):
        """Returns the following model:

            Y = Add(MatMul(Transpose(X), Transpose(W)), B)

        Where:
        - Transpose(X) is applied only if `transA=True`
        - Transpose(W) is applied only if `transB=True`
        - W and B can be graph inputs or initializers
        """
        tape = ir.tape.Tape()
        inputs = []
        bias_shape = weight_shape[0] if transB else weight_shape[-1]
        output_shape = ir.Shape(("?",) * input_shape.rank())

        x = ir.val("X", shape=input_shape, type=ir.TensorType(ir.DataType.FLOAT))

        if weight_as_inputs:
            w = ir.val("W", shape=weight_shape, type=ir.TensorType(ir.DataType.FLOAT))
            inputs.append(w)
        else:
            w = ir.tensor(
                self.rng.uniform(-0.5, 0.5, weight_shape).astype("float32"), name="W"
            )
            w = tape.initializer(w)

        if bias_as_inputs:
            b = ir.val(
                "B", shape=ir.Shape([bias_shape]), type=ir.TensorType(ir.DataType.FLOAT)
            )
            inputs.append(b)
        else:
            b = ir.tensor(self.rng.uniform(-0.5, 0.5, bias_shape).astype("float32"), name="B")
            b = tape.initializer(b)

        x_t, w_t = None, None
        if transA:
            x_t = tape.op("Transpose", inputs=[x], attributes={"perm": permA})

        if transB:
            w_t = tape.op("Transpose", inputs=[w], attributes={"perm": permB})

        y = tape.op("MatMul", inputs=[x_t if transA else x, w_t if transB else w])
        y = tape.op(
            "Add",
            inputs=[y, b],
            output=ir.val("Y", shape=output_shape, type=ir.TensorType(ir.DataType.FLOAT)),
        )

        # Build the model
        ir_model = ir.Model(
            ir.Graph(
                inputs=[x, *inputs],
                outputs=[y],
                nodes=tape.nodes,
                initializers=tape.initializers,
                opset_imports={"": 20},
                name="test_model",
            ),
            ir_version=10,
        )
        onnx_checker.CheckerPass(True)(ir_model)
        ir_model = shape_inference.infer_shapes(ir_model)
        return ir_model

    def check_matmul_add_to_gemm_incompatible_shapes(self, **kwargs):
        base_model = self.get_test_model(**kwargs)

        updated_model = self.clone_model(base_model)
        tracer = MatchingTracer()
        count = _matmul_add_to_gemm.matmul_add_to_gemm_rule.apply_to_model(
            updated_model, tracer=tracer
        )

        # Check that the model is unchanged
        self.assertEqual(count, 0)

        # Check that the error message is the expected one
        tracer_match = tracer.best_matches_map[_matmul_add_to_gemm.matmul_add_to_gemm_rule][0]
        self.assertEqual(tracer_match.status.value, MatchStatus.CONDITION_FAILED)
        self.assertRegex(
            tracer_match.match_result.reason, "Rank of input_a and input_b must be 2"
        )


class MatMulAddToGemmTest(_MatMulAddToGemmTestBase):
    @parameterized.expand(
        [
            ("initializers", False, False),
            ("inputs", True, True),
        ]
    )
    def test_matmul_add_to_gemm(self, _, weight_as_inputs, bias_as_inputs):
        base_model = self.get_test_model(
            input_shape=ir.Shape((512, 256)),
            weight_shape=ir.Shape((256, 64)),
            weight_as_inputs=weight_as_inputs,
            bias_as_inputs=bias_as_inputs,
        )
        updated_model = self.clone_model(base_model)
        count = _matmul_add_to_gemm.rules.apply_to_model(updated_model)

        # Check MatMul + Add are fused into Gemm
        self.assertEqual(count, 1)
        self.assertEqual(len(updated_model.graph), 1)

        # Prepare inputs
        if weight_as_inputs and bias_as_inputs:
            inputs = (
                self.rng.random((512, 256), dtype=np.float32),
                self.rng.random((256, 64), dtype=np.float32),
                self.rng.random((64), dtype=np.float32),
            )
        else:
            inputs = (self.rng.random((512, 256), dtype=np.float32),)

        # Check inference
        testing.assert_numerically_equal(base_model, updated_model, inputs)

        # Validate serialized model
        output_model_proto = ir.serde.serialize_model(updated_model)
        onnx.checker.check_model(output_model_proto, full_check=True)

    def test_matmul_add_to_gemm_incompatible_shapes(self):
        kwargs = {
            "input_shape": ir.Shape((1, 256, 512)),
            "weight_shape": ir.Shape((1, 512, 64)),
        }
        return super().check_matmul_add_to_gemm_incompatible_shapes(**kwargs)


class TransAMatMulAddToGemmTest(_MatMulAddToGemmTestBase):
    @parameterized.expand(
        [
            ("initializers", False, False),
            ("inputs", True, True),
        ]
    )
    def test_transpose_a_matmul_add_to_gemm(self, _, weight_as_inputs, bias_as_inputs):
        base_model = self.get_test_model(
            input_shape=ir.Shape((256, 512)),
            weight_shape=ir.Shape((256, 64)),
            weight_as_inputs=weight_as_inputs,
            bias_as_inputs=bias_as_inputs,
            transA=True,
        )
        updated_model = self.clone_model(base_model)
        count = _matmul_add_to_gemm.rules.apply_to_model(updated_model)

        # Check MatMul(Transpose, W) + Add are fused into Gemm
        self.assertEqual(count, 1)
        self.assertEqual(len(updated_model.graph), 1)

        # Prepare inputs
        if weight_as_inputs and bias_as_inputs:
            inputs = (
                self.rng.random((256, 512), dtype=np.float32),
                self.rng.random((256, 64), dtype=np.float32),
                self.rng.random((64,), dtype=np.float32),
            )
        else:
            inputs = (self.rng.random((256, 512), dtype=np.float32),)

        # Check inference
        testing.assert_numerically_equal(base_model, updated_model, inputs)

        # Validate serialized model
        output_model_proto = ir.serde.serialize_model(updated_model)
        onnx.checker.check_model(output_model_proto, full_check=True)

    def test_transpose_a_matmul_add_to_gemm_incompatible_shapes(self):
        kwargs = {
            "input_shape": ir.Shape((1, 256, 512)),
            "weight_shape": ir.Shape((1, 256, 64)),
            "transA": True,
            "permA": [0, 2, 1],
        }
        return super().check_matmul_add_to_gemm_incompatible_shapes(**kwargs)


class TransBMatMulAddToGemmTest(_MatMulAddToGemmTestBase):
    @parameterized.expand(
        [
            ("initializers", False, False),
            ("inputs", True, True),
        ]
    )
    def test_transpose_b_matmul_add_to_gemm(self, _, weight_as_inputs, bias_as_inputs):
        base_model = self.get_test_model(
            input_shape=ir.Shape((512, 256)),
            weight_shape=ir.Shape((64, 256)),
            weight_as_inputs=weight_as_inputs,
            bias_as_inputs=bias_as_inputs,
            transB=True,
        )
        updated_model = self.clone_model(base_model)
        count = _matmul_add_to_gemm.rules.apply_to_model(updated_model)

        # Check MatMul(X, Transpose) + Add are fused into Gemm
        self.assertEqual(count, 1)
        self.assertEqual(len(updated_model.graph), 1)

        # Prepare inputs
        if weight_as_inputs and bias_as_inputs:
            inputs = (
                self.rng.random((512, 256), dtype=np.float32),
                self.rng.random((64, 256), dtype=np.float32),
                self.rng.random((64,), dtype=np.float32),
            )
        else:
            inputs = (self.rng.random((512, 256), dtype=np.float32),)

        # Check inference
        testing.assert_numerically_equal(base_model, updated_model, inputs)

        # Validate serialized model
        output_model_proto = ir.serde.serialize_model(updated_model)
        onnx.checker.check_model(output_model_proto, full_check=True)

    def test_transpose_b_matmul_add_to_gemm_incompatible_shapes(self):
        kwargs = {
            "input_shape": ir.Shape((1, 512, 256)),
            "weight_shape": ir.Shape((1, 64, 256)),
            "transB": True,
            "permB": [0, 2, 1],
        }
        return super().check_matmul_add_to_gemm_incompatible_shapes(**kwargs)


class TransABMatMulAddToGemmTest(_MatMulAddToGemmTestBase):
    @parameterized.expand(
        [
            ("initializers", False, False),
            ("inputs", True, True),
        ]
    )
    def test_transpose_ab_matmul_add_to_gemm(self, _, weight_as_inputs, bias_as_inputs):
        base_model = self.get_test_model(
            input_shape=ir.Shape((256, 512)),
            weight_shape=ir.Shape((64, 256)),
            weight_as_inputs=weight_as_inputs,
            bias_as_inputs=bias_as_inputs,
            transA=True,
            transB=True,
        )
        updated_model = self.clone_model(base_model)
        count = _matmul_add_to_gemm.rules.apply_to_model(updated_model)

        # Check MatMul(Transpose, Transpose) + Add are fused into Gemm
        self.assertEqual(count, 1)
        self.assertEqual(len(updated_model.graph), 1)

        # Prepare inputs
        if weight_as_inputs and bias_as_inputs:
            inputs = (
                self.rng.random((256, 512), dtype=np.float32),
                self.rng.random((64, 256), dtype=np.float32),
                self.rng.random((64), dtype=np.float32),
            )
        else:
            inputs = (self.rng.random((256, 512), dtype=np.float32),)

        # Check inference
        testing.assert_numerically_equal(base_model, updated_model, inputs)

        # Validate serialized model
        output_model_proto = ir.serde.serialize_model(updated_model)
        onnx.checker.check_model(output_model_proto, full_check=True)

    def test_transpose_ab_matmul_add_to_gemm_incompatible_shapes(self):
        kwargs = {
            "input_shape": ir.Shape((1, 256, 512)),
            "weight_shape": ir.Shape((1, 64, 256)),
            "transA": True,
            "transB": True,
            "permA": [0, 2, 1],
            "permB": [0, 2, 1],
        }
        return super().check_matmul_add_to_gemm_incompatible_shapes(**kwargs)


if __name__ == "__main__":
    unittest.main()



# Copyright (c) Microsoft Corporation.
# Licensed under the MIT License.
"""Fuses successive Min/Max patterns in ONNX graphs.

Supported transformations:
- Min(Min(X, c1, c2, ...), d1, d2, ...) → Min(X, fused_const)
- Max(Max(X, c1, c2, ...), d1, d2, ...) → Max(X, fused_const)
- Min(Max(X, lb1, lb2, ...), ub1, ub2, ...) → Clip(X, lb, ub)
- Max(Min(X, ub1, ub2, ...), lb1, lb2, ...) → Clip(X, lb, ub)

Where:
    - fused_const is the reduction (min or max) over all constant inputs.
    - For Clip fusion:
        * All constant inputs must be scalars.
        * The effective lower bound is the maximum of all lower-bound constants.
        * The effective upper bound is the minimum of all upper-bound constants.

        For the case of Max(Min(X, upper_bound), lower_bound):
            * The rule applies only if lower_bound ≤ upper_bound.

General constraints:
    - The first input may be any tensor.
    - All other inputs must be constant tensors (from Constant nodes or initializers).
"""

import abc
import functools
from typing import ClassVar

import numpy as np
import onnx_ir as ir

from onnxscript.rewriter._basics import MatchResult
from onnxscript.rewriter._rewrite_rule import RewriteRuleClassBase, RewriteRuleSet


class _FuseMinMaxBase(RewriteRuleClassBase, abc.ABC):
    """Base class for Min/Max fusion rewrites.

    Constraints:
        - All inputs except the first must be constants (from Constant nodes or initializers).
        - If ``need_scalars`` is True (Clip fusion), all constants must be scalars.
        - If ``check_bounds`` is True (Clip fusion in the pattern Max(Min(X, upper_bound), lower_bound)), lower_bound ≤ upper_bound.
    """

    need_scalars: ClassVar = False
    check_bounds: ClassVar = False

    @abc.abstractmethod
    def compute_constants(
        self,
        first_node: ir.Node,
        second_node: ir.Node,
        input_name: str = "",
    ) -> list[tuple[ir.Tensor, str]]: ...

    def rewrite(self, op, x, out1, out2):
        first_node = out1.producer()
        second_node = out2.producer()

        # Compute new constants for the fused op
        constants = self.compute_constants(first_node, second_node, x.name)

        initializers = [op.initializer(constant, name=name) for constant, name in constants]

        return op.op(
            self.op_type,
            inputs=[x, *initializers],
        )

    def _is_scalar(self, v: np.ndarray) -> bool:
        return np.isscalar(v) or np.size(v) == 1

    def check(self, context, out1, out2, **_):
        """Condition to check if we need to replace the pattern.

        Conditions:
            - The min and max input nodes must not be graph inputs.
            - These inputs (except the first) must be constant values (from Constant nodes or initializers).
            - In the case of Min(Max) and Max(Min) patterns:
                * All inputs must be scalars (as Clip requires scalars).
                For Max(Min) pattern:
                    * The lower bound must be less than or equal to the upper bound.

        Returns:
            MatchResult:
                Success if we need to replace the pattern, Failure otherwise.
        """
        del context  # Not used
        check_result = MatchResult()

        first_node = out1.producer()
        second_node = out2.producer()

        # Ensure all inputs except the first are constants
        for input_ in first_node.inputs[1:] + second_node.inputs[1:]:
            if ir.convenience.get_const_tensor(input_) is None:
                return check_result.fail(f"{input_.name} is not a constant.")

            # If scalars are required (Clip fusion), enforce scalar-ness
            if self.need_scalars and not self._is_scalar(input_.const_value.numpy()):
                return check_result.fail(f"{input_.name} is not a scalar.")

        if self.need_scalars and self.check_bounds:
            # For Clip fusion in the case of Max(Min(X, upper_bound), lower_bound): check that lower_bound <= upper_bound
            lower_bound, upper_bound = self.compute_constants(first_node, second_node)
            if lower_bound[0].numpy() > upper_bound[0].numpy():
                return check_result.fail(
                    f"Invalid bounds: lower bound ({lower_bound[0].numpy()}) is greater "
                    f"than upper bound ({upper_bound[0].numpy()})."
                )

        return check_result


class FuseSuccessiveMin(_FuseMinMaxBase):
    """Replaces ``Min(Min(X, c1, c2, ...), d1, d2, ...)`` with ``Min(X, fused_const)``.

    Constraints:
        - All inputs except the first must be constants (from Constant nodes or initializers).
    """

    op_type: ClassVar = "Min"

    def compute_constants(
        self,
        first_node: ir.Node,
        second_node: ir.Node,
        input_name: str = "",
    ) -> list[tuple[ir.Tensor, str]]:
        inputs = first_node.inputs[1:] + second_node.inputs[1:]
        values = [input_.const_value.numpy() for input_ in inputs]
        return [(ir.tensor(functools.reduce(np.minimum, values)), f"{input_name}_min")]

    def pattern(self, op, x):
        return op.Min(
            op.Min(x, _allow_other_inputs=True, _outputs=["out1"]),
            _allow_other_inputs=True,
            _outputs=["out2"],
        )


class FuseSuccessiveMax(_FuseMinMaxBase):
    """Replaces ``Max(Max(X, c1, c2, ...), d1, d2, ...)`` with ``Max(X, fused_const)``.

    Constraints:
        - All inputs except the first must be constants (from Constant nodes or initializers).
    """

    op_type: ClassVar = "Max"

    def compute_constants(
        self,
        first_node: ir.Node,
        second_node: ir.Node,
        input_name: str = "",
    ) -> list[tuple[ir.Tensor, str]]:
        inputs = first_node.inputs[1:] + second_node.inputs[1:]
        values = [input_.const_value.numpy() for input_ in inputs]
        return [(ir.tensor(functools.reduce(np.maximum, values)), f"{input_name}_max")]

    def pattern(self, op, x):
        return op.Max(
            op.Max(x, _allow_other_inputs=True, _outputs=["out1"]),
            _allow_other_inputs=True,
            _outputs=["out2"],
        )


class FuseMaxMinToClip(_FuseMinMaxBase):
    """Replaces ``Min(Max(X, lb1, lb2, ...), ub1, ub2, ...)`` with ``Clip(X, lb, ub)``.

    Constraints:
        - All inputs except the first must be constants (from Constant nodes or initializers).
        - All constant inputs must be scalars.
        - The effective lower bound is ``max(lb1, lb2, ...)``.
        - The effective upper bound is ``min(ub1, ub2, ...)``.
    """

    op_type: ClassVar = "Clip"
    need_scalars: ClassVar = True

    def compute_constants(
        self,
        first_node: ir.Node,
        second_node: ir.Node,
        input_name: str = "",
    ) -> list[tuple[ir.Tensor, str]]:
        lower_bound = np.max([input_.const_value.numpy() for input_ in first_node.inputs[1:]])
        upper_bound = np.min([input_.const_value.numpy() for input_ in second_node.inputs[1:]])
        return [
            (ir.tensor(lower_bound), f"{input_name}_min"),
            (ir.tensor(upper_bound), f"{input_name}_max"),
        ]

    def pattern(self, op, x):
        return op.Min(
            op.Max(x, _allow_other_inputs=True, _outputs=["out1"]),
            _allow_other_inputs=True,
            _outputs=["out2"],
        )


class FuseMinMaxToClip(_FuseMinMaxBase):
    """Replaces ``Max(Min(X, ub1, ub2, ...), lb1, lb2, ...)`` with ``Clip(X, lb, ub)``.

    Constraints:
        - All inputs except the first must be constants (from Constant nodes or initializers).
        - All constant inputs must be scalars.
        - The effective lower bound is ``max(lb1, lb2, ...)``.
        - The effective upper bound is ``min(ub1, ub2, ...)``.
        - Requires ``lower_bound <= upper_bound``.
    """

    op_type: ClassVar = "Clip"
    need_scalars: ClassVar = True
    check_bounds: ClassVar = True

    def compute_constants(
        self,
        first_node: ir.Node,
        second_node: ir.Node,
        input_name: str = "",
    ) -> list[tuple[ir.Tensor, str]]:
        upper_bound = np.min([input_.const_value.numpy() for input_ in first_node.inputs[1:]])
        lower_bound = np.max([input_.const_value.numpy() for input_ in second_node.inputs[1:]])
        return [
            (ir.tensor(lower_bound), f"{input_name}_min"),
            (ir.tensor(upper_bound), f"{input_name}_max"),
        ]

    def pattern(self, op, x):
        return op.Max(
            op.Min(x, _allow_other_inputs=True, _outputs=["out1"]),
            _allow_other_inputs=True,
            _outputs=["out2"],
        )


min_min_rule = FuseSuccessiveMin().rule()
max_max_rule = FuseSuccessiveMax().rule()
min_max_rule = FuseMinMaxToClip().rule()
max_min_rule = FuseMaxMinToClip().rule()


rules = RewriteRuleSet(
    [
        min_min_rule,
        max_max_rule,
        min_max_rule,
        max_min_rule,
    ]
)



# Copyright (c) Microsoft Corporation.
# Licensed under the MIT License.
from __future__ import annotations

import unittest

import onnx_ir as ir

import onnxscript
from onnxscript.rewriter.rules.fusion import _rms_normalization


class RmsNormOnnxFusionsTest(unittest.TestCase):
    def test_rms_normalization_fusion(self):
        opset23 = onnxscript.values.Opset("", 23)

        @onnxscript.script()
        def rms_norm_script(embedding, layernorm_weight):
            two = opset23.Constant(value_float=2.0)
            pow_1 = opset23.Pow(embedding, two)
            mean = opset23.ReduceMean(pow_1, [-1], keepdims=1, noop_with_empty_axes=0)
            epsilon = opset23.Constant(value_float=1e-05)
            add_1 = opset23.Add(mean, epsilon)
            val_244 = opset23.Sqrt(add_1)
            rsqrt = opset23.Reciprocal(val_244)
            mul_3 = opset23.Mul(embedding, rsqrt)
            mul_4 = opset23.Mul(layernorm_weight, mul_3)
            return mul_4

        rms_norm_model_proto = rms_norm_script.to_model_proto(
            input_types=[onnxscript.FLOAT[128], onnxscript.FLOAT[128]],
            output_types=[onnxscript.FLOAT[128]],
        )
        model = ir.serde.deserialize_model(rms_norm_model_proto)
        count = _rms_normalization.fuse_rms_normalization(model)
        self.assertEqual(count, 1)
        self.assertEqual(model.graph.node(-1).op_type, "RMSNormalization")


if __name__ == "__main__":
    unittest.main()



# Copyright (c) Microsoft Corporation.
# Licensed under the MIT License.
from __future__ import annotations

from typing import Union

import onnx_ir as ir

import onnxscript.rewriter._fusion_utils as _fusion_utils
from onnxscript.rewriter import _basics, pattern

Dim = Union[int, ir.SymbolicDim]


class OnnxGroupQueryAttention(pattern.RewriteRuleClassBase):
    def __init__(self):
        super().__init__("ONNXGQA", remove_nodes=False)

    def pattern(
        self,
        op,
        query_BHSD,
        key_BHkvSD,
        value_BHkvSD,
        past_key_BHkvSpD,
        past_value_BHkvSpD,
    ):
        # Concatenate past_key cache and current key, expand across heads
        # that share key/value.

        present_key_BHkvStD = op.Concat(past_key_BHkvSpD, key_BHkvSD, axis=-2)
        present_key_BHkv1StD = op.Unsqueeze(present_key_BHkvStD, 2)
        present_key_BHkvGStD = op.Expand(present_key_BHkv1StD, pattern.ANY_VALUE)
        present_key_BHStD = op.Reshape(
            present_key_BHkvGStD, pattern.ANY_VALUE, _outputs=["present_key_BHStD"]
        )

        # Concatenate past_value cache and current value, expand across heads
        # that share key/value.
        present_value_BHkvStD = op.Concat(past_value_BHkvSpD, value_BHkvSD, axis=-2)
        present_value_BHkv1StD = op.Unsqueeze(present_value_BHkvStD, 2)
        present_value_BHkvGStD = op.Expand(present_value_BHkv1StD, pattern.ANY_VALUE)
        present_value_BHStD = op.Reshape(
            present_value_BHkvGStD, pattern.ANY_VALUE, _outputs=["present_value_BHStD"]
        )

        attention_BHSDh = op.Attention(
            query_BHSD,
            present_key_BHStD,
            present_value_BHStD,
            pattern.Var("mask", can_match_none=True),
            _outputs=["attention_BHSDh"],
        )

        return attention_BHSDh, present_key_BHkvStD, present_value_BHkvStD

    def check(
        self,
        context: _basics.MatchContext,
        query_BHSD,
        key_BHkvSD,
        value_BHkvSD,
        past_key_BHkvSpD,
        past_value_BHkvSpD,
        present_key_BHStD,
        present_value_BHStD,
        **_,
    ):
        bindings: dict[str, Dim] = {}
        # Check that inputs to new Attention node have expected shapes
        _fusion_utils.check_shape(bindings, query_BHSD, ["B", "H", "S", "D"])
        _fusion_utils.check_shape(bindings, key_BHkvSD, ["B", "Hkv", "S", "D"])
        _fusion_utils.check_shape(bindings, value_BHkvSD, ["B", "Hkv", "S", "D"])
        _fusion_utils.check_shape(bindings, past_key_BHkvSpD, ["B", "Hkv", "P", "D"])
        _fusion_utils.check_shape(bindings, past_value_BHkvSpD, ["B", "Hkv", "P", "D"])
        # We need to check that the Expand/Reshape arguments are as expected.
        # As a substitute, we check that the outputs of Expand=>Reshape have expected shapes.
        # TODO (rama): May be better to check the actual Expand/Reshape arguments.
        _fusion_utils.check_shape(bindings, present_key_BHStD, ["B", "H", "S+P", "D"])
        _fusion_utils.check_shape(bindings, present_value_BHStD, ["B", "H", "S+P", "D"])

        return True

    def rewrite(
        self,
        op,
        query_BHSD,
        key_BHkvSD,
        value_BHkvSD,
        past_key_BHkvSpD,
        past_value_BHkvSpD,
        mask,
        attention_BHSDh,
        **_,
    ):
        original_attention_node = attention_BHSDh.producer()
        original_attrs = original_attention_node.attributes
        return op.Attention(
            query_BHSD,
            key_BHkvSD,
            value_BHkvSD,
            mask,
            past_key_BHkvSpD,
            past_value_BHkvSpD,
            **original_attrs,
            _outputs=3,
        )


_basic_gqa_rule = OnnxGroupQueryAttention.rule()

gqa_rules = pattern.RewriteRuleSet([_basic_gqa_rule])

fuse_gqa = _fusion_utils.apply_fusion_rules(gqa_rules)



# Copyright (c) Microsoft Corporation.
# Licensed under the MIT License.
from __future__ import annotations

from typing import Sequence, Union

import onnx_ir as ir

from onnxscript.rewriter import _fusion_utils, _ir_utils, pattern

Dim = Union[int, ir.SymbolicDim]


class PackedQKVForGQAFusion(pattern.RewriteRuleClassBase):
    def __init__(self):
        super().__init__("PackedQKVForGQA", remove_nodes=False)

    def pattern(
        self,
        op,
        packed_qkv,
        past_key,
        past_value,
        seqlens_k,
        total_seq_length,
        cos,
        sin,
        q_num_heads,
        kv_num_heads,
        interleaved,
        start1,
        end1,
        start2,
        end2,
        start3,
        end3,
    ):
        """Pattern to detect sliced Q, K, V passed to GQA and replace with packed QKV."""

        # Slice packed QKV into query, key, and value
        query_BSD = op.Slice(packed_qkv, start1, end1, [2], [1], _outputs=["query_sliced"])
        key_BSDkv = op.Slice(packed_qkv, start2, end2, [2], [1], _outputs=["key_sliced"])
        value_BSDkv = op.Slice(packed_qkv, start3, end3, [2], [1], _outputs=["value_sliced"])

        # Pass sliced Q, K, V to GroupQueryAttention
        return op.GroupQueryAttention(
            query_BSD,
            key_BSDkv,
            value_BSDkv,
            past_key,
            past_value,
            seqlens_k,
            total_seq_length,
            cos,
            sin,
            # mask, # TODO: this is not a valid input for GQA
            num_heads=q_num_heads,
            kv_num_heads=kv_num_heads,
            do_rotary=1,
            rotary_interleaved=interleaved,
            # skipped optional attributes: local_window_size, scale, smooth_softmax, softcap
            _domain="com.microsoft",
            _outputs=3,
        )

    def check(
        self,
        op,
        packed_qkv,
        query_sliced,
        key_sliced,
        value_sliced,
        q_num_heads,
        kv_num_heads,
        start1,
        end1,
        start2,
        end2,
        start3,
        end3,
        **_,
    ):
        check_result = pattern.MatchResult()
        self.bindings: dict[str, Dim] = {}

        def no_match(val: ir.Value, dims: Sequence[str]) -> bool:
            return not _fusion_utils.check_shape_bool(self.bindings, val, dims)

        # Check that if x is being split into q, k, v correctly
        # based on hidden sizes
        if packed_qkv is None or packed_qkv.shape is None or len(packed_qkv.shape) != 3:
            return check_result.fail("packed_qkv is not a 3D tensor.", packed_qkv)
        hidden_size = packed_qkv.shape[2]
        if not isinstance(hidden_size, int):
            return check_result.fail("Hidden size is not an integer.", packed_qkv)
        q_nh = q_num_heads.value
        kv_nh = kv_num_heads.value
        if not isinstance(q_nh, int) or not isinstance(kv_nh, int):
            return check_result.fail(
                "Could not determine the number of heads for query, key and value.",
            )
        head_size = hidden_size // (q_nh + (2 * kv_nh))
        q_hidden_size = head_size * q_nh
        kv_hidden_size = head_size * kv_nh
        if not (
            _ir_utils.is_singleton_value(start1, 0)
            and _ir_utils.is_singleton_value(end1, q_hidden_size)
            and _ir_utils.is_singleton_value(start2, q_hidden_size)
            and _ir_utils.is_singleton_value(end2, (q_hidden_size + kv_hidden_size))
            and _ir_utils.is_singleton_value(start3, (q_hidden_size + kv_hidden_size))
            and _ir_utils.is_singleton_value(end3, lambda x: x >= hidden_size)
        ):
            return check_result.fail(
                "packed_qkv is not being split into q, k, v correctly based on hidden sizes.",
                packed_qkv,
            )

        # Check packed_qkv shape (B, S, D)
        if no_match(packed_qkv, ["B", "S", "D"]):
            return check_result.fail(
                f"Shape mismatch: {packed_qkv} does not match expected dimensions ['B', 'S', 'D']",
                packed_qkv,
            )

        # Check query, key, and value shapes (B, S, Dh)
        if no_match(query_sliced, ["B", "S", "Dq"]):
            return check_result.fail(
                f"Shape mismatch: {query_sliced} does not match expected dimensions ['B', 'S', 'Dq']",
                query_sliced,
            )
        if no_match(key_sliced, ["B", "S", "Dkv"]):
            return check_result.fail(
                f"Shape mismatch: {key_sliced} does not match expected dimensions ['B', 'S', 'Dkv']",
                key_sliced,
            )
        if no_match(value_sliced, ["B", "S", "Dkv"]):
            return check_result.fail(
                f"Shape mismatch: {value_sliced} does not match expected dimensions ['B', 'S', 'Dkv']",
                value_sliced,
            )

        # Ensure Dh = Dg + 2*Dkv
        D = self.bindings.get("D")
        Dq = self.bindings.get("Dq")
        Dkv = self.bindings.get("Dkv")

        if not isinstance(D, int) or not isinstance(Dq, int) or not isinstance(Dkv, int):
            return check_result.fail(
                "Could not determine the hidden sizes of query, key, and value.",
            )

        if Dq + (2 * Dkv) != D:  # type: ignore[operator]
            return check_result.fail(
                f"Hidden size of query, key and value do not add up to hidden size: {D} != {Dq} + (2 * {Dkv})",
            )

        return True

    def rewrite(
        self,
        op,
        packed_qkv,
        past_key,
        past_value,
        seqlens_k,
        total_seq_length,
        cos,
        sin,
        q_num_heads,
        kv_num_heads,
        interleaved,
        **_,
    ):
        """Rewrite the sliced Q, K, V into a packed QKV MatMul input for GQA."""

        # Pass packed QKV directly to GroupQueryAttention
        return op.GroupQueryAttention(
            packed_qkv,
            None,
            None,
            past_key,
            past_value,
            seqlens_k,
            total_seq_length,
            cos,
            sin,
            num_heads=q_num_heads,
            kv_num_heads=kv_num_heads,
            do_rotary=1,
            rotary_interleaved=interleaved,
            _domain="com.microsoft",
            _outputs=3,
        )


# Define the fusion rule
packed_qkv_for_gqa_rule = PackedQKVForGQAFusion.rule()

# Add the rule to the GQA rewrite rule set
fuse_qkv_gqa_rules = pattern.RewriteRuleSet([packed_qkv_for_gqa_rule])

# Apply the fusion rules
fuse_qkv_gqa = _fusion_utils.apply_fusion_rules(fuse_qkv_gqa_rules)



# Copyright (c) Microsoft Corporation.
# Licensed under the MIT License.
import unittest

import numpy as np

from onnxscript import ir
from onnxscript.rewriter import rewrite, testing
from onnxscript.rewriter.rules.common import (
    affine_conv_fusion_rule,
    conv_affine_fusion_rule,
)


class FuseConvAffineTest(unittest.TestCase):
    def clone_model(self, model: ir.Model) -> ir.Model:
        return ir.from_proto(ir.to_proto(model))

    def test_conv_affine_fusion(self):
        tape = ir.tape.Tape()
        x = ir.val("x", dtype=ir.DataType.FLOAT, shape=ir.Shape([1, 3, 32, 32]))
        w = tape.initializer(ir.tensor(np.ones((3, 3, 3, 3), dtype=np.float32), name="w"))
        b = tape.initializer(ir.tensor(np.ones((3,), dtype=np.float32), name="b"))
        scale = tape.initializer(ir.tensor(np.array([2.0], dtype=np.float32), name="scale"))
        offset = tape.initializer(ir.tensor(np.array([3.0], dtype=np.float32), name="offset"))

        conv_out = tape.op("Conv", [x, w, b], attributes={"pads": [1, 1, 1, 1]})
        mul_out = tape.op("Mul", [conv_out, scale])
        z = tape.op(
            "Add",
            [mul_out, offset],
            output=ir.val(
                "z",
                dtype=ir.DataType.FLOAT,
                shape=ir.Shape([1, 3, 32, 32]),
            ),
        )

        model = ir.Model(
            ir.Graph(
                inputs=[x],
                outputs=[z],
                nodes=tape.nodes,
                initializers=tape.initializers,
                opset_imports={"": 17},
            ),
            ir_version=8,
        )
        rewritten_model = self.clone_model(model)
        rewritten_model = rewrite(
            rewritten_model,
            pattern_rewrite_rules=[conv_affine_fusion_rule],
        )
        # Check that Mul and Add are fused into Conv
        self.assertEqual(model.graph.num_nodes() - 2, rewritten_model.graph.num_nodes())

        # Check that the results are numerically equal
        rng = np.random.default_rng(42)
        inputs = [
            rng.random((1, 3, 32, 32), dtype=np.float32),
        ]
        testing.assert_numerically_equal(model, rewritten_model, inputs)

    def test_affine_conv_fusion_without_pad(self):
        tape = ir.tape.Tape()
        x = ir.val("x", dtype=ir.DataType.FLOAT, shape=ir.Shape([1, 3, 32, 32]))
        w = tape.initializer(ir.tensor(np.ones((3, 3, 3, 3), dtype=np.float32), name="w"))
        b = tape.initializer(ir.tensor(np.ones((3,), dtype=np.float32), name="b"))
        scale = tape.initializer(ir.tensor(np.array([2.0], dtype=np.float32), name="scale"))
        offset = tape.initializer(ir.tensor(np.array([3.0], dtype=np.float32), name="offset"))

        mul_out = tape.op("Mul", [x, scale])
        z = tape.op(
            "Add",
            [mul_out, offset],
            output=ir.val(
                "z",
                dtype=ir.DataType.FLOAT,
                shape=ir.Shape([1, 3, 32, 32]),
            ),
        )
        conv_out = tape.op("Conv", [z, w, b], attributes={"pads": [0, 0, 0, 0]})

        model = ir.Model(
            ir.Graph(
                inputs=[x],
                outputs=[conv_out],
                nodes=tape.nodes,
                initializers=tape.initializers,
                opset_imports={"": 17},
            ),
            ir_version=8,
        )
        rewritten_model = self.clone_model(model)
        rewritten_model = rewrite(
            rewritten_model,
            pattern_rewrite_rules=[affine_conv_fusion_rule],
        )
        # Check that Mul and Add are fused into Conv
        self.assertEqual(model.graph.num_nodes() - 2, rewritten_model.graph.num_nodes())

        # Check that the results are numerically equal
        rng = np.random.default_rng(42)
        inputs = [
            rng.random((1, 3, 32, 32), dtype=np.float32),
        ]
        testing.assert_numerically_equal(model, rewritten_model, inputs)


if __name__ == "__main__":
    unittest.main()



# Copyright (c) Microsoft Corporation.
# Licensed under the MIT License.

from __future__ import annotations

import numpy as np
import onnx
import onnx_ir as ir


def _get_numpy_value(
    val: ir.Value | None, dtype: ir.DataType | None = None, size_limit: int | None = None
) -> np.ndarray | None:
    """Returns the numpy value of a constant value, if available.

    It returns None if the value is not a constant value, or if the value is not of
    the specified element dtype, or if the size of the value exceeds the specified
    size_limit.
    """
    if val is None:
        return None
    const_value = val.const_value
    if const_value is not None:
        if dtype is not None and const_value.dtype != dtype:
            return None
        if size_limit is not None and const_value.size > size_limit:
            return None
        try:
            # Turn the constant value into a numpy array representation with the
            # specifics of this conversion handled by the tensor type
            array = const_value.numpy()
            # Can/should not reinterpret strings via .view, resulting in
            #   "TypeError: Cannot change data-type for array of references."
            # There is also no reason to reinterpret strings, this is only
            # relevant for some arithmetic types
            if const_value.dtype != ir.DataType.STRING:
                # Reinterpret the array with `.view()` because some
                # implementations  of ir.TensorProtocol (e.g. PyTorch<=2.7) do
                # not use ml_dtypes for bfloat16 etc.
                array = array.view(const_value.dtype.numpy())
        except FileNotFoundError:
            # External data is not available.
            # logger.warning(
            #     "External data for value '%s' is not available. "
            #     "This may lead to incorrect constant folding.",
            #     val.name,
            # )
            return None
        assert isinstance(array, np.ndarray)
        return array
    return None


def _do_onnx_inference(node: ir.Node) -> None:
    """Run ONNX shape and type inference on a single node, updating its outputs in place."""

    def get_constant_value(x: ir.Value) -> onnx.TensorProto | None:
        value = _get_numpy_value(x, size_limit=20)
        if value is not None:
            assert x.const_value is not None
            return ir.serde.serialize_tensor(x.const_value)
        return None

    def get_type(index: int, value: ir.Value) -> onnx.TypeProto:
        if value.type is None:
            raise ValueError(
                f"Type of input {index} value {value.name} of node {node.name} not known"
            )
        type_proto = ir.serde.serialize_type(value.type)
        if value.shape is not None:
            ir.serde.serialize_shape_into(type_proto, value.shape)
        return type_proto

    input_types = {x.name: get_type(i, x) for i, x in enumerate(node.inputs) if x is not None}
    input_data = {x.name: get_constant_value(x) for x in node.inputs if x is not None}
    input_data = {k: v for k, v in input_data.items() if v is not None}

    # TODO: pass in constant values, ir_version
    schema = onnx.defs.get_schema(node.op_type, node.version, node.domain)
    output_types = onnx.shape_inference.infer_node_outputs(
        schema,
        ir.serde.serialize_node(node),
        input_types,  # type: ignore[arg-type]
        input_data,  # type: ignore[arg-type]
    )
    for output in node.outputs:
        if output.name in output_types:
            inferred_type = output_types[output.name]
            # TODO: merge types, check for conflicts
            inferred_shape = ir.serde.deserialize_type_proto_for_shape(inferred_type)
            # NOTE: forward shape inference
            output.merge_shapes(inferred_shape)
            output.type = ir.serde.deserialize_type_proto_for_type(inferred_type)


def infer_outputs(node: ir.Node) -> None:
    try:
        _do_onnx_inference(node)
    except Exception as e:  # pylint: disable=broad-exception-caught
        # Inference is best-effort: failures (missing schema, unknown types,
        # serialization errors, etc.) are recorded but must not crash the builder.
        # TODO: compose with any existing error
        node.metadata_props["pkg.onnxscript.shape_inference_error"] = str(e)



# Copyright (c) Microsoft Corporation.
# Licensed under the MIT License.
from __future__ import annotations

import onnx_ir as ir

from onnxscript.rewriter import _fusion_utils, _ir_utils, pattern

"""
RMS Normalization: This is referred to as SimplifiedLayerNormalization in the ORT codebase.
See https://github.com/microsoft/onnxruntime/blob/6d9636f07cccdb6e4ac453087ad54c3bc9854d50/onnxruntime/core/graph/contrib_ops/contrib_defs.cc#L2981

Key points for the fusion optimization:
* Input and scale are allowed to be of different types.
* The normalization of the input can be done in a different precision than the input type,
which is also the precision of reciprocal_rms returned by operation.
* Input (x) must be: float or double or float16 or bfloat16
* Scale must be: float or double or float16 or bfloat16
* Normalization precision must be float or double
"""

float_types = frozenset(
    [
        ir.DataType.FLOAT,
        ir.DataType.FLOAT16,
        ir.DataType.BFLOAT16,
        ir.DataType.DOUBLE,
    ]
)
fp_float_types = frozenset([ir.DataType.FLOAT, ir.DataType.DOUBLE])


class RmsNormFusion(pattern.RewriteRuleClassBase):
    def __init__(self, name: str, _mul_order: bool):
        super().__init__(name)
        self._mul_order = _mul_order

    def pattern(self, op, x, scale, epsilon, compute_dtype, target_dtype):
        x = pattern.OrValue([op.Cast(x, to=compute_dtype), x])
        x_square = op.Pow(x, 2.0)
        mean_square = op.ReduceMean(x_square, [-1], keepdims=1, noop_with_empty_axes=0)
        mean_square_plus_epsilon = op.Add(mean_square, epsilon)
        rms = op.Sqrt(mean_square_plus_epsilon)
        reciprocal_rms = op.Reciprocal(rms)
        normalized = op.Mul(x, reciprocal_rms)
        normalized = pattern.OrValue([op.Cast(normalized, to=target_dtype), normalized])
        # To support float16, we need to ensure the scale is casted or not.
        scale = pattern.OrValue([op.Cast(scale, to=compute_dtype), scale])
        # Workaround: can't use OrValue for final (returned) value
        if self._mul_order:
            return op.Mul(normalized, scale)
        else:
            return op.Mul(scale, normalized)

    def check(
        self, op, x, scale, epsilon, compute_dtype, target_dtype, **_
    ) -> pattern.MatchResult:  # type: ignore[name-defined]
        """Check if the pattern matches conditions for use of SimplifiedLayerNormalization op."""
        check_result = pattern.MatchResult()
        # epsilon must be a scalar
        epsilon_value = _ir_utils.get_singleton_value(epsilon)
        if not isinstance(epsilon_value, float):  # TODO: support other types
            return check_result.fail("Epsilon is not a float value.", epsilon)
        if x.dtype not in float_types:
            return check_result.fail("Input is not a float type.", x)
        if scale.dtype not in float_types:
            return check_result.fail("Scale is not a float type.", scale)
        self._stash_dtype = compute_dtype.as_int() if compute_dtype is not None else x.dtype
        if self._stash_dtype not in fp_float_types:
            return check_result.fail("Normalization precision is not a float or double type.")
        # target_dtype is guaranteed to be the same as scale type in a well-typed input
        # for Mul(scale, normalized) to work. There is no need to check it here for a well-typed input.
        # TODO (rama): Consider adding checks to protect against incorrectly typed models:
        return check_result

    def rewrite(self, op, x, scale, epsilon, **_):
        # Note: ORT's SimplifiedLayerNormalization was placed in onnx domain by mistake.
        # No need to use com.microsoft domain here; but this is a custom op in ORT.
        return op.SimplifiedLayerNormalization(
            x,
            scale,
            axis=-1,
            epsilon=_ir_utils.get_singleton_value(epsilon),
            stash_type=self._stash_dtype,
        )


_rule1 = RmsNormFusion.rule("RmsNormFusion1", _mul_order=False)
_rule2 = RmsNormFusion.rule("RmsNormFusion2", _mul_order=True)

rms_normalization_rules = [_rule1, _rule2]
rms_normalization_ruleset = pattern.RewriteRuleSet(rms_normalization_rules)


fuse_rms_normalization = _fusion_utils.apply_fusion_rules(rms_normalization_ruleset)



# Copyright (c) Microsoft Corporation.
# Licensed under the MIT License.
"""Stable APIs for PyTorch 2.7."""

from __future__ import annotations

__all__ = [
    "check_model",
    "convert_version",
    "get_torchlib_ops",
    "optimize",
    "save_model_with_external_data",
]

from onnxscript._framework_apis.torch_2_6 import (
    check_model,
    convert_version,
    get_torchlib_ops,
    optimize,
    save_model_with_external_data,
)



# Copyright (c) Microsoft Corporation.
# Licensed under the MIT License.
"""Fusion optimizations for ORT backend."""

__all__ = [
    "optimize_for_ort",
    "ORT_PATTERN_REWRITE_RULES",
]


from onnxscript.rewriter.ort_fusions._core import ORT_PATTERN_REWRITE_RULES, optimize_for_ort



# Copyright (c) Microsoft Corporation.
# Licensed under the MIT License.
"""Does the following transformation:
- Div(Clip(Add(x))) -> HardSigmoid
- Mul(HardSigmoid(x), x) -> HardSwish
"""

from __future__ import annotations

import numpy as np
import onnx_ir as ir

from onnxscript.rewriter import pattern
from onnxscript.rewriter._basics import MatchResult
from onnxscript.rewriter._ir_utils import is_singleton_value
from onnxscript.rewriter._rewrite_rule import RewriteRuleSet


class _HardSigmoidFusionBase(pattern.RewriteRuleClassBase):
    """HardSwish requires constant values so we check in base class."""

    def check(
        self,
        op,
        x: ir.Value,
        clip_min: ir.Value,
        clip_max: ir.Value,
        bias: ir.Value,
        divisor: ir.Value,
    ) -> MatchResult:
        check_result = MatchResult()

        if not is_singleton_value(clip_min, 0.0, rtol=1e-4):
            return check_result.fail("Swish requires min value of 0 for clip")
        if not is_singleton_value(clip_max, 6.0, rtol=1e-4):
            return check_result.fail("Swish requires max value of 6 for clip")
        if not is_singleton_value(bias, 3.0, rtol=1e-4):
            return check_result.fail("Swish requires bias value of 3")
        if not is_singleton_value(divisor, 6.0, rtol=1e-4):
            return check_result.fail("Swish requires divisor value of 6")
        return check_result


class HardSwishFusion(_HardSigmoidFusionBase):
    """Fuse Add(_, 3) + Clip<0, 6>(_) + Mul + Div(_, 6) into HardSwish

    In this case we can't make HardSigmoid fusion first. The Mul
    is placed before Div while HardSigmoid requires Add+Clip+Div.
    """

    def pattern(
        self,
        op,
        x: ir.Value,
        clip_min: ir.Value,
        clip_max: ir.Value,
        bias: ir.Value,
        divisor: ir.Value,
    ) -> ir.Value:
        out = op.Clip(x + bias, clip_min, clip_max) * x
        out = out / divisor
        return out

    def rewrite(
        self,
        op,
        x: ir.Value,
        clip_min: ir.Value,
        clip_max: ir.Value,
        bias: ir.Value,
        divisor: ir.Value,
    ) -> ir.Value:
        return op.HardSwish(x)


class HardSwishFusionFromHardSigmoid(pattern.RewriteRuleClassBase):
    """Fuse HardSigmoid<alpha=1/6, beta=0.5> + Mul into HardSwish"""

    def pattern(self, op, x: ir.Value) -> ir.Value:
        # Floating point matching for 1/6 is not exact, so we use isclose below
        out = op.HardSigmoid(x, _allow_other_attributes=True, _outputs=["hardsigmoid_out"])
        out = out * x
        return out

    def check(self, op, x: ir.Value, hardsigmoid_out: ir.Value) -> MatchResult:
        check_result = MatchResult()
        hardsigmoid = hardsigmoid_out.producer()
        # Use getter to protect when 'alpha' / 'beta' is not in attributes
        alpha = hardsigmoid.attributes.get_float("alpha", -1)
        beta = hardsigmoid.attributes.get_float("beta", -1)
        if not np.isclose(alpha, 1 / 6):
            return check_result.fail(
                "HardSigmoid alpha must be 1/6 to get fused into HardSwish"
            )
        if not np.isclose(beta, 0.5):
            return check_result.fail(
                "HardSigmoid beta must be 0.5 to get fused into HardSwish"
            )
        return check_result

    def rewrite(self, op, x: ir.Value, hardsigmoid_out: ir.Value) -> ir.Value:
        return op.HardSwish(x)


class HardSigmoidFusion(_HardSigmoidFusionBase):
    """Fuse HardSigmoid only for HardSwish hyper-parameters: alpha=1/6, beta=0.5"""

    def pattern(
        self,
        op,
        x: ir.Value,
        clip_min: ir.Value,
        clip_max: ir.Value,
        bias: ir.Value,
        divisor: ir.Value,
    ) -> ir.Value:
        out = op.Clip(x + bias, clip_min, clip_max)
        out = out / divisor
        return out

    def rewrite(
        self,
        op,
        x: ir.Value,
        clip_min: ir.Value,
        clip_max: ir.Value,
        bias: ir.Value,
        divisor: ir.Value,
    ) -> ir.Value:
        return op.HardSigmoid(x, alpha=1 / 6, beta=0.5)


def fuse_hardswish_rules() -> RewriteRuleSet:
    """Returns the rewrite rules for fusing HardSwish and HardSigmoid."""
    return RewriteRuleSet(
        [
            HardSwishFusion().rule(),
            HardSigmoidFusion().rule(),
            HardSwishFusionFromHardSigmoid().rule(),
        ],
        commute=True,
    )
